# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import click
from tqdm.auto import tqdm
from tsbench.config import DATASET_REGISTRY
from tsbench.constants import DEFAULT_DATA_PATH
from tsbench.evaluations.aws import default_session
from tsbench.evaluations.aws.s3 import upload_directory
from ._main import datasets


@datasets.command(short_help="Upload locally available datasets to S3.")
@click.option(
    "--bucket",
    required=True,
    help="The S3 bucket to upload the dataset to.",
)
@click.option(
    "--path",
    type=str,
    default=DEFAULT_DATA_PATH,
    show_default=True,
    help="The local path where the datasets are stored.",
)
@click.option(
    "--prefix",
    type=str,
    default="data",
    show_default=True,
    help="The prefix in the S3 bucket.",
)
def upload(path: str, bucket: str, prefix: str):
    """
    Uploads the data for all datasets available locally to an S3 bucket.
    """
    s3 = default_session().client("s3")
    for config in tqdm(DATASET_REGISTRY.values()):
        upload_directory(
            s3, path / config.name(), bucket, f"{prefix}/{config.name()}"
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import pickle
import random

import torch
import numpy as np

import pandas as pd
from tqdm.auto import tqdm

from typing import Iterator
from src.gluonts.dataset.common import DataEntry, Dataset, ListDataset


PREDICTION_INPUT_NAMES = [
    "feat_static_cat",
    "feat_static_real",
    "past_time_feat",
    "past_target_cdf",
    "past_observed_values",
    "future_time_feat",
    "past_is_pad",
]


class Params:
    """
    Class that loads hyperparameters from a json file.

    Example:
    params = Params(json_path)
    print(params.learning_rate)
    params.learning_rate = 0.5  # change the value of learning_rate in params
    """

    def __init__(self, json_path: str = None):
        if json_path is not None:
            with open(json_path) as f:
                params = json.load(f)
                self.__dict__.update(params)

    def save(self, json_path):
        with open(json_path, "w") as f:
            json.dump(self.__dict__, f, indent=4, ensure_ascii=False)

    def update(self, json_path):
        """
        Loads parameters from json file.
        """
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    @property
    def dict(self):
        """
        Gives dict-like access to Params instance by
        params.dict['learning_rate']
        """
        return self.__dict__


class AttackResults:
    def __init__(self, batch, perturbation, true_future_target, attack_idx):
        self.batch = batch
        self.perturbation = perturbation
        self.true_future_target = true_future_target
        self.attack_idx = attack_idx


class Metrics:
    def __init__(
        self,
        mse,
        mape,
        ql,
        quantiles=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
    ):
        self.mse = mse
        self.mape = mape
        self.ql = ql
        self.quantiles = quantiles


def smoothed_inference(
    batch,
    past_target,
    net,
    sigma,
    device,
    num_noised_samples: int = 100,
    intermediate_noise: float = None,
    retain_positivity: bool = True,
):
    outputs = []

    for _ in tqdm(range(num_noised_samples), leave=False):
        noised_past_target = change_device(past_target, device)
        noised_past_target += torch.normal(
            mean=torch.zeros(noised_past_target.shape, device=device),
            std=sigma * torch.abs(noised_past_target),
        )
        if retain_positivity:
            noised_past_target = torch.clamp(noised_past_target, min=0)

        noised_inputs = {key: batch[key] for key in PREDICTION_INPUT_NAMES}
        noised_inputs["past_target_cdf"] = noised_past_target

        sample = net(**noised_inputs)

        outputs.append(sample.detach().cpu().numpy())

        del noised_past_target, noised_inputs, sample
        torch.cuda.empty_cache()

    return np.concatenate(outputs, axis=1)


def requires_grad_(model: torch.nn.Module, requires_grad: bool) -> None:
    for param in model.parameters():
        param.requires_grad_(requires_grad)


def get_augmented_dataset(dataset, num_noises: int = 100, sigma: float = 0.1):
    train_data_list = list(iter(dataset.train_ds))
    train_length = len(train_data_list)

    for _ in range(num_noises):
        for idx in range(train_length):
            target = train_data_list[idx]["target"]
            data = {
                "start": train_data_list[idx]["start"],
                "target": target
                + np.random.normal(
                    loc=np.zeros_like(target), scale=sigma * target
                ),
                "feat_static_cat": train_data_list[idx][
                    "feat_static_cat"
                ].copy(),
                "item_id": None,
                "source": None,
            }
            train_data_list.append(data)

    random.shuffle(train_data_list)
    return ListDataset(
        data_iter=train_data_list, freq=dataset.freq, one_dim_target=False
    )


def add_ts_dataframe(
    data_iterator: Iterator[DataEntry], freq
) -> Iterator[DataEntry]:
    for data_entry in data_iterator:
        data = data_entry.copy()

        index = pd.date_range(
            start=data["start"],
            freq=freq,
            periods=data["target"].shape[-1],
        )
        data["ts"] = pd.DataFrame(index=index, data=data["target"].transpose())
        yield data


def ts_iter(dataset: Dataset, freq) -> pd.DataFrame:
    for data_entry in add_ts_dataframe(iter(dataset), freq):
        yield data_entry["ts"]


def load_pickle(filename):
    with open(filename, "rb") as f:
        while True:
            try:
                yield pickle.load(f)
            except EOFError:
                break


def change_device(var, device):
    if isinstance(var, torch.Tensor):
        if var.device == "cpu":
            var.to(device)
            return var
        elif var.device != device:
            return var.cpu().to(device)
        else:
            return var
    if isinstance(var, dict):
        for key in var.keys():
            if isinstance(var[key], torch.Tensor):
                var[key] = var[key].to(device)
        return var

    return torch.from_numpy(var).float().to(device)


def quantile_loss(true, pred, quantile):
    denom = np.abs(true).sum(1).sum(1)  # batch x 1
    if true.ndim < 3:
        true = true.reshape(true.shape + tuple([1] * (3 - true.ndim)))
    if pred.ndim < 3:
        pred = pred.reshape(pred.shape + tuple([1] * (3 - pred.ndim)))
    batch, time, dim = true.shape
    ql = np.zeros(batch)
    idx = 0
    for y_hat, y in zip(pred, true):
        num = 0
        for t in range(time):
            for j in range(dim):
                num += (
                    (1 - quantile) * abs(y_hat[t, j] - y[t, j])
                    if y_hat[t, j] > y[t, j]
                    else quantile * abs(y_hat[t, j] - y[t, j])
                )
        ql[idx] = num
        idx += 1
    denom[denom == 0] = 1
    if (denom != 0).any():
        return (2 * ql / denom).reshape(batch, 1, 1)
    else:
        return None


def calc_loss(
    attack_data,
    forecasts,
    attack_idx,
    target_items,
    quantiles=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
):
    testset_size = sum(
        [
            attack_data[i].true_future_target.shape[0]
            for i in range(len(attack_data))
        ]
    )
    mse = {
        key: np.zeros((testset_size, len(attack_idx), len(target_items)))
        for key in forecasts.keys()
    }
    mape = {
        key: np.zeros((testset_size, len(attack_idx), len(target_items)))
        for key in forecasts.keys()
    }
    ql = {
        key: np.zeros((len(quantiles), testset_size, 1, len(target_items)))
        for key in forecasts.keys()
    }
    testset_idx = 0

    for i in tqdm(range(len(attack_data))):
        true_future_target = attack_data[i].true_future_target
        batch_size = true_future_target.shape[0]

        for attack_type in forecasts.keys():
            if (
                true_future_target[:, attack_idx][..., target_items] != 0
            ).prod() == 0:
                mape[attack_type][testset_idx : testset_idx + batch_size] = (
                    np.abs(
                        forecasts[attack_type][i][:, :, attack_idx][
                            ..., target_items
                        ].mean(1)
                        - true_future_target[:, attack_idx][..., target_items]
                    )
                )
                mse[attack_type][testset_idx : testset_idx + batch_size] = (
                    forecasts[attack_type][i][:, :, attack_idx][
                        ..., target_items
                    ].mean(1)
                    - true_future_target[:, attack_idx][..., target_items]
                ) ** 2
                for j, quantile in enumerate(quantiles):
                    pred = np.quantile(
                        a=forecasts[attack_type][i][
                            :, :, attack_idx, target_items
                        ],
                        q=quantile,
                        axis=1,
                    )
                    true = true_future_target[:, attack_idx][..., target_items]
                    ql[attack_type][
                        j, testset_idx : testset_idx + batch_size
                    ] = quantile_loss(true, pred, quantile)
            else:
                mape[attack_type][testset_idx : testset_idx + batch_size] = (
                    np.abs(
                        forecasts[attack_type][i][:, :, attack_idx][
                            ..., target_items
                        ].mean(1)
                        / true_future_target[:, attack_idx][..., target_items]
                        - 1
                    )
                )
                mse[attack_type][testset_idx : testset_idx + batch_size] = (
                    mape[attack_type][testset_idx : testset_idx + batch_size]
                    ** 2
                )
                for j, quantile in enumerate(quantiles):
                    pred = np.quantile(
                        a=forecasts[attack_type][i][:, :, attack_idx][
                            ..., target_items
                        ],
                        q=quantile,
                        axis=1,
                    )
                    true = true_future_target[:, attack_idx][..., target_items]
                    ql[attack_type][
                        j, testset_idx : testset_idx + batch_size
                    ] = quantile_loss(true, pred, quantile)
        testset_idx += batch_size

    return mse, mape, ql



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

from typing import cast, Optional
from pathlib import Path
import os
import random
import click
import pytorch_lightning as pl
import pytorch_lightning.loggers as pll
import torch
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks import ModelCheckpoint

from meta.models.module import MetaLightningModule
from meta.models.common import load_weights
from meta.models import get_model, MODEL_REGISTRY
from meta.datasets import get_data_module, DATA_MODULE_REGISTRY
from meta.metrics import CRPS, QuantileLoss, QuantileWidth
from meta.callbacks.common import get_save_dir_from_csvlogger
from meta.callbacks import (
    ForecastPlotLoggerCallback,
    ParameterCountCallback,
    InitialSaveCallback,
    ForecastSupportSetAttentionPlotLoggerCallback,
    LossPlotLoggerCallback,
    CheatLossPlotLoggerCallback,
    MacroCRPSPlotCallback,
)
from experiments.evaluation.eval_real_cheat import large_training_evaluation


@click.command()
@click.option(
    "--data_dir",
    required=True,
    default=(
        cast(str, os.getenv("SM_CHANNEL_DATASETS"))
        if os.getenv("SM_INPUT_DIR") is not None
        else Path.home() / ".mxnet" / "gluon-ts"
    ),
)
@click.option(
    "--output_dir",
    required=True,
    default=os.getenv("SM_MODEL_DIR"),
)
@click.option("--seed", default=0)
# Experiment setup
@click.option(
    "--model_name",
    required=True,
    help="The model to fit on the data and to use for predictions.",
    type=click.Choice(MODEL_REGISTRY.keys()),
)
@click.option("--quantiles", default="0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9")
@click.option(
    "--dm_name", required=True, type=click.Choice(DATA_MODULE_REGISTRY.keys())
)
@click.option("--support_set_size", default=3)
@click.option("--standardize", default=True)
# ------  Data module specific hyperparameters ------
# super data module
@click.option("--dm_super_dataset_names_train", default="electricity")
@click.option("--dm_super_dataset_names_val", default="electricity")
@click.option("--dm_super_dataset_names_test", default="electricity")
@click.option(
    "--dm_super_dataset_sampling",
    default="weighted",
    type=click.Choice(["weighted", "uniform"]),
)
@click.option("--dm_super_catch22_train", default=False)
@click.option("--dm_super_catch22_val_test", default=False)
@click.option("--dm_super_cheat", default=0.0)
@click.option("--dm_super_num_queries", default=1)
# artificial data module
@click.option("--dm_artificial_dataset_name", default="marker")
@click.option("--dm_artificial_prediction_length", default=24)
@click.option("--dm_artificial_train_set_size", default=10000)
# artifical cheat data module
@click.option("--dm_cheat_train_set_size", default=10000)
@click.option("--dm_cheat_val_set_size", default=500)
@click.option("--dm_cheat_test_set_size", default=500)
@click.option("--dm_cheat_prediction_length", default=24)
@click.option("--dm_cheat_cheat_chance", default=1.0)
@click.option("--dm_cheat_noise_level", default=0.0)
@click.option("--dm_cheat_query_length_scale", default=5.0)
@click.option("--dm_cheat_support_length_scale_same_as_query", default=True)
# artifical cheat counterfactual data module
@click.option("--dm_cheat_counterfactual_train_set_size", default=10000)
@click.option("--dm_cheat_counterfactual_val_set_size", default=500)
@click.option("--dm_cheat_counterfactual_test_set_size", default=500)
@click.option("--dm_cheat_counterfactual_prediction_length", default=24)
@click.option("--dm_cheat_counterfactual_cheat_chance", default=1.0)
@click.option("--dm_cheat_counterfactual_noise_level", default=0.0)
@click.option("--dm_cheat_counterfactual_query_length_scale", default=5.0)
@click.option("--dm_cheat_counterfactual_counterfactual_size", default=1)
@click.option("--dm_cheat_counterfactual_counterfactual_mixing", default=False)
@click.option(
    "--dm_cheat_counterfactual_support_length_scale_same_as_query",
    default=True,
)
# Common hyperparameters
@click.option("--pretrained", default=None)
@click.option("--learning_rate", default=1e-3)
@click.option("--lr_on_plateau_patience", default=5)
@click.option("--lr_on_plateau_factor", default=0.3)
@click.option("--context_length_multiple", default=4)
@click.option("--support_length_multiple", default=4)
@click.option("--model_prediction_length", default=0)
@click.option("--max_epochs", default=400)
@click.option("--max_training_time", default="02:00:00:00")
@click.option("--num_batches", default=1024)
@click.option("--batch_size_train", default=1000)
@click.option("--batch_size_val_test", default=1000)
@click.option("--accumulate_grad_batches", default=0)
@click.option("--early_stopping_min_delta", default=0.001)
@click.option("--early_stopping_patience", default=20)
@click.option(
    "--monitor_metric",
    required=True,
    help="This metric is used for EarlyStopping, ModelCheckpoint, LRScheduler callbacks.",
)
@click.option("--n_logging_samples", default=5)
@click.option("--log_plot_every_n_epochs", default=10)
@click.option("--num_workers", default=8)
# ------  Model specific hyperparameters ------
# lstm feedforward model
@click.option("--lstm_feedforward_num_lstm_layers", default=1)
@click.option("--lstm_feedforward_query_out_channels", default=32)
@click.option("--lstm_feedforward_decoder_hidden_size", default=32)
# CNN Iwata model
@click.option("--cnn_iwata_num_heads", default=1)
@click.option("--cnn_iwata_supps_kernel_size", default=3)
@click.option("--cnn_iwata_supps_out_channels", default=32)
# Iwata model
@click.option("--iwata_num_heads", default=1)
@click.option("--iwata_supps_out_channels", default=32)
@click.option("--iwata_query_out_channels", default=32)
@click.option("--iwata_supps_bidirectional", default=True)
@click.option("--iwata_decoder_hidden_size", default=32)
@click.option("--iwata_lstm_num_layers", default=1)
# TCN model
@click.option("--tcn_num_heads", default=1)
@click.option("--tcn_num_channels", default=64)
@click.option("--tcn_kernel_size", default=2)
@click.option("--tcn_num_layers", default=5)
@click.option("--tcn_hidden_size", default=64)
@click.option("--tcn_decoder_hidden_size", default=64)
def main(
    data_dir: str,
    output_dir: str,
    seed: int,
    # Experiment setup
    model_name: str,
    quantiles: str,
    dm_name: str,
    support_set_size: int,
    standardize: bool,
    # Common hyperparameters
    pretrained: str,
    learning_rate: float,
    lr_on_plateau_patience: float,
    lr_on_plateau_factor: float,
    context_length_multiple: int,
    support_length_multiple: int,
    model_prediction_length: Optional[int],
    max_epochs: int,
    max_training_time: str,
    num_batches: int,
    batch_size_train: int,
    batch_size_val_test: int,
    accumulate_grad_batches: Optional[int],
    early_stopping_min_delta: float,
    early_stopping_patience: int,
    monitor_metric: str,
    n_logging_samples: int,
    log_plot_every_n_epochs: int,
    num_workers: int,
    # Model hyperparameters
    **kwargs: int,
):
    """
    Trains a model in the meta learning framework.
    """
    args_to_save = locals()
    random.seed(seed)
    torch.manual_seed(seed)
    pl.seed_everything(seed, workers=True)
    kwargs["dm_super_seed"] = seed

    quantiles = quantiles.split(",")

    # -------------------- Configure the data loader ---------------------------------------------------
    dm = get_data_module(
        dm_name,
        support_set_size=support_set_size,
        support_length_multiple=support_length_multiple,
        standardize=standardize,
        data_dir=Path(data_dir),
        batch_size_train=batch_size_train,
        batch_size_val_test=batch_size_val_test,
        context_length_multiple=context_length_multiple,
        num_workers=num_workers,
        **{
            key[len(dm_name) + 1 :]: value
            for key, value in kwargs.items()
            if key.startswith(dm_name)
        },
    )
    dm.setup()
    # -------------------- Initialize the model ---------------------------------------------------
    model = get_model(
        model_name,
        prediction_length=model_prediction_length or dm.prediction_length,
        quantiles=quantiles,
        **{
            key[len(model_name) + 1 :]: value
            for key, value in kwargs.items()
            if key.startswith(model_name)
        },
    )
    if pretrained:
        model = load_weights(model=model, path_to_weights=pretrained)

    module = MetaLightningModule(
        model,
        loss=QuantileLoss(quantiles),
        crps=CRPS(quantiles, rescale=dm.standardize),
        lr_scheduler_monitor=monitor_metric,
        crps_scaled=CRPS(quantiles) if dm.standardize else None,
        quantile_width=QuantileWidth(quantiles),
        lr=learning_rate,
        lr_on_plateau_patience=lr_on_plateau_patience,
        lr_on_plateau_factor=lr_on_plateau_factor,
        val_dataset_names=dm.dataset_names_val,
        test_dataset_names=dm.dataset_names_test,
    )

    # -------------------- configure callbacks ---------------------------------------------------
    # for the plotting callback we need to get some validation data
    log_batch_train, log_batch_val = dm.get_log_batches(n_logging_samples)

    # early stopping callback
    early_stop_callback = EarlyStopping(
        monitor=monitor_metric,
        min_delta=early_stopping_min_delta,
        patience=early_stopping_patience,
        verbose=True,
        mode="min",
    )

    checkpoint_callback = ModelCheckpoint(
        monitor=monitor_metric,
        filename=monitor_metric + "--{epoch:02d}--{train_loss:.2f}",
        save_top_k=1,
        mode="min",
        save_last=True,
    )

    callbacks = [
        ParameterCountCallback(),
        InitialSaveCallback(args_to_save=args_to_save),
        early_stop_callback,
        checkpoint_callback,
    ]
    if dm_name == "dm_super":
        callbacks.append(
            LossPlotLoggerCallback(every_n_epochs=log_plot_every_n_epochs)
        )
        if len(kwargs["dm_super_dataset_names_val"].split(",")) > 1:
            callbacks.append(
                MacroCRPSPlotCallback(every_n_epochs=log_plot_every_n_epochs)
            )
    elif dm_name == "dm_cheat":
        callbacks.append(
            CheatLossPlotLoggerCallback(
                dataset_names_val=dm.dataset_names_val,
                every_n_epochs=log_plot_every_n_epochs,
            )
        )

    # add callback that only works with attention
    attention_models = ["iwata", "cnn_iwata", "tcn"]
    if model_name in attention_models:
        callbacks.extend(
            [
                ForecastSupportSetAttentionPlotLoggerCallback(
                    log_batch_train,
                    quantiles=quantiles,
                    split="train",
                    every_n_epochs=log_plot_every_n_epochs,
                ),
                ForecastSupportSetAttentionPlotLoggerCallback(
                    log_batch_val,
                    quantiles=quantiles,
                    split="val",
                    every_n_epochs=log_plot_every_n_epochs,
                ),
            ]
        )
    else:
        callbacks.extend(
            [
                ForecastPlotLoggerCallback(
                    log_batch_val,
                    quantiles=quantiles,
                    split="val",
                    every_n_epochs=log_plot_every_n_epochs,
                ),
                ForecastPlotLoggerCallback(
                    log_batch_train,
                    quantiles=quantiles,
                    split="train",
                    every_n_epochs=log_plot_every_n_epochs,
                ),
            ]
        )

    # -------------------- train model ---------------------------------------------------
    trainer = pl.Trainer(
        gpus=1,
        log_every_n_steps=1,
        limit_train_batches=num_batches,
        val_check_interval=1.0,
        max_epochs=max_epochs,
        max_time=max_training_time,
        logger=pll.CSVLogger(output_dir),
        accumulate_grad_batches=accumulate_grad_batches or None,
        callbacks=callbacks,
        # profiler="pytorch",
    )

    trainer.fit(module, dm)
    trainer.test(module, dm, ckpt_path="best")

    # -------------------- evaluate model and store results ---------------------------------------------------
    # load weights with best validation loss during training period
    model_best = load_weights(
        model=model, path_to_weights=checkpoint_callback.best_model_path
    )
    if dm_name == "dm_super" and model_name in attention_models:
        save_dir = get_save_dir_from_csvlogger(trainer.logger)
        large_training_evaluation(
            save_dir=save_dir,
            model_name=model_name,
            seed=seed,
            support_length_multiple=support_length_multiple,
            dm_super_cheat=kwargs["dm_super_cheat"],
            datasets=kwargs["dm_super_dataset_names_test"].split(","),
            dm_super_catch22_train=kwargs["dm_super_catch22_train"],
            quantiles=quantiles,
            model=model_best,
            data_dir=Path(data_dir),
        )


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Type, TypeVar, Any
import torch.nn as nn
from lightkit.utils import get_generic_type
from lightkit.nn import Configurable
from dataclasses import MISSING

MODEL_REGISTRY: Dict[str, Type[nn.Module]] = {}


M = TypeVar("M", bound=Type[nn.Module])


def register_model(cls: M) -> M:
    """
    Registers the provided class in the global model registry.
    """
    MODEL_REGISTRY[cls.name()] = cls
    return cls


def get_model(name: str, **kwargs: Any) -> nn.Module:
    """
    This method creates the model configuration of the model with the specified
    name. The provided keyword arguments must contain ALL arguments required by
    the model configuration. Superfluous arguments may be provided and are
    simply ignored.

    In case the name is unknown or parameters for the model config's initializer are missing, an
    assertion error occurs.

    Parameters
    ----------
    name: str
        The canonical name of the model. See `MODEL_REGISTRY`.
    kwargs: Any
        Keyword arguments passed to the initializer of the model config.

    Returns
    -------
    ModelConfig
        The model configuration.
    """
    # Get the model
    assert name in MODEL_REGISTRY, f"Model name '{name}' is unknown."
    model_cls = MODEL_REGISTRY[name]
    config_cls = get_generic_type(model_cls, Configurable)

    # Filter the required parameters
    all_params = set(key for key in config_cls.__dataclass_fields__.keys())
    required = {
        key
        for key, field in config_cls.__dataclass_fields__.items()
        if (field.default is MISSING and field.default_factory is MISSING)  # type: ignore
    }
    assert all(
        r in kwargs for r in required
    ), "Keyword arguments missing at least one parameter."

    return model_cls(
        config_cls(**{k: v for k, v in kwargs.items() if k in all_params})
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pickle
import numpy as np
import math


def create_table(path):
    objects = []
    if "adv" in path:
        types = "adv"
    elif "_rs" in path:
        types = "rs"
    elif "rt" in path:
        types = "da"
    else:
        types = "no_defense"
    print(types)
    with open(path, "rb") as openfile:
        while True:
            try:
                objects.append(pickle.load(openfile))
            except EOFError:
                break

    mse = []
    mape = []
    wql = []
    sparsity = []

    result = objects[0]
    batch = len(result.mse["1"])
    c = 1.96 / math.sqrt(batch)

    for result in objects:
        print("mse loss:")
        for key in result.mse.keys():
            sparsity.append(key)
            print(
                key,
                np.asarray(result.mse[key]).mean(),
                "+-",
                np.asarray(result.mse[key]).std() * c,
            )
            mse.append(
                (
                    np.asarray(result.mse[key]).mean(),
                    np.asarray(result.mse[key]).std() * c,
                )
            )

        print("mape loss:")
        for key in result.mape.keys():
            print(
                key,
                np.asarray(result.mape[key]).mean(),
                "+-",
                np.asarray(result.mape[key]).std() * c,
            )
            mape.append(
                (
                    np.asarray(result.mape[key]).mean(),
                    np.asarray(result.mape[key]).std() * c,
                )
            )

        print("wQL:")
        for key in result.ql.keys():
            print(
                key,
                np.asarray(result.ql[key]).mean(),
                "+-",
                np.asarray(result.ql[key]).std() * c,
            )
            wql.append(
                (
                    np.asarray(result.ql[key]).mean(),
                    np.asarray(result.ql[key]).std() * c,
                )
            )

    with open("table_" + types + ".txt", "w") as f:
        for i in range(len(mse)):
            # with std
            print(
                sparsity[i]
                + " & {:.4f}$\pm${:.4f} & {:.4f}$\pm${:.4f} & {:.4f}$\pm${:.4f}\\\\".format(
                    mape[i][0],
                    mape[i][1],
                    mse[i][0],
                    mse[i][1],
                    wql[i][0],
                    wql[i][1],
                ),
                file=f,
            )
            # without std
            # print(sparsity[i] + ' & {:.4f} & {:.4f} & {:.4f}\\\\'.format(mape[i][0], mse[i][0], wql[i][0]), file=f)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .config import Config, EnsembleConfig
from .dataset import DATASET_REGISTRY, DatasetConfig
from .model import MODEL_REGISTRY, ModelConfig, TrainConfig

__all__ = [
    "Config",
    "DATASET_REGISTRY",
    "DatasetConfig",
    "EnsembleConfig",
    "MODEL_REGISTRY",
    "ModelConfig",
    "TrainConfig",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
from typing import Dict, Any
from pytorch_lightning import Callback, Trainer, LightningModule

from .common import get_save_dir_from_csvlogger


class InitialSaveCallback(Callback):  # type: ignore
    """
    This callback saves the initial model using the save method of the LightKit
    ConfigModule. This allows to load the model later without access to the
    hyper-parameters needed to instantiate the class. Additionally, a
    dictionary of arguments is stored.

    Args:
        args_to_save: Contains the arguments that are stored.
    """

    def __init__(self, args_to_save: Dict[str, Any]) -> None:
        super().__init__()
        self.args_to_save = args_to_save

    def on_pretrain_routine_start(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        base_dir = get_save_dir_from_csvlogger(trainer.logger)
        # save the dictionary
        with open(base_dir / "args.json", "w") as fp:
            json.dump(self.args_to_save, fp)

        # save model
        save_dir = base_dir / "initial_model"
        save_dir.mkdir(parents=True, exist_ok=True)
        pl_module.model.save(path=save_dir)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import torch
from torch import nn
from meta.data.batch import SeriesBatch


class FeatureExtractor(nn.Module, ABC):
    """
    Base class for feature extractors.
    """

    @abstractmethod
    def forward(self, series: SeriesBatch) -> SeriesBatch:
        """
        Computes features for each time point for each time series.

        Parameters
        ----------
        series
            A SeriesBatch containing sequences of size [batch, sequence length, n_input],
            where n_input is typically 1 + * which corresponds to the univariate time series itself
            and additional time features, e.g. relative time.

        Returns
        -------
        SeriesBatch
            The transformed batch with sequences of size [batch, sequence length, n_features].
        """


class IdentityFeatureExtractor(FeatureExtractor):
    """
    Dummy feature extractor which implements the identity function.
    """

    def __init__(self):
        super().__init__()
        self.feature_extractor = nn.Identity()

    def forward(self, series: SeriesBatch) -> torch.Tensor:
        return self.feature_extractor(series)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional
import torch
from torch.nn.utils.rnn import pad_sequence

from .sampling import Triplet
from .dataset import TimeSeries
from meta.common.torch import tensor_to_np


@dataclass
class SeriesBatch:
    """
    A batch of series from different base datasets, represented by the padded
    batch, the lengths of the series and the splits sizes indicating the
    corresponding base dataset.
    """

    sequences: (
        torch.Tensor
    )  # shape [batch, num_sequences, max_sequence_length]
    lengths: torch.Tensor  # shape [batch]
    split_sections: torch.Tensor  # shape [batch]
    scales: Optional[torch.Tensor] = (
        None  # shape[batch, 2] contains mean and std the ts has been scaled with
    )

    @classmethod
    def from_lists(
        cls, lists: List[List[TimeSeries]], squeeze: bool = False
    ) -> SeriesBatch:
        """
        Initializes a series batch from the provided series.

        Args:
            lists: List of lists of time series.
            squeeze: If the last dimension should be squeezed. This should be done for future queries
                which do not contain co-variates and are, thus, univariate

        Returns:
            The initialized series batch.
        """
        split_sections = [len(section) for section in lists]
        series = [s for sublist in lists for s in sublist]
        if squeeze:
            values = [torch.squeeze(s.values, -1) for s in series]
        else:
            values = [s.values for s in series]
        return SeriesBatch(
            pad_sequence(values, batch_first=True),
            lengths=torch.as_tensor([len(s) for s in series]),
            split_sections=torch.as_tensor(split_sections),
            scales=(
                torch.stack([s.scale for s in series])
                if series[0].scale is not None
                else None
            ),
        )

    def pin_memory(self):
        self.sequences = self.sequences.pin_memory()
        self.lengths = self.lengths.pin_memory()
        self.split_sections = self.split_sections.pin_memory()
        if self.scales is not None:
            self.scales = self.scales.pin_memory()
        return self

    def unpad(self, to_numpy: bool = False, squeeze: bool = False):
        """
        Unpad sequences in batch by using the lengths.
        """
        t = tensor_to_np if to_numpy else lambda x: x
        sq = torch.squeeze if squeeze else lambda x: x

        splits = torch.split(self.sequences, self.split_sections.tolist())
        lengths = torch.split(self.lengths, self.split_sections.tolist())
        out = []
        for l_split, split in zip(lengths, splits):
            un_padded = [t(sq(s[:l])) for l, s in zip(l_split, split)]
            if squeeze:
                out.extend(un_padded)
            else:
                out.append(un_padded)
        return out

    def to(self, device) -> SeriesBatch:
        return SeriesBatch(
            self.sequences.to(device),
            self.lengths.to(device),
            self.split_sections.to(device),
            scales=self.scales.to(device) if self.scales is not None else None,
        )

    def rescale(self) -> SeriesBatch:
        """
        Redo standardization.

        The series must contain the same time series in the same order as the
        dataset.
        """
        m = self.scales[:, 0].unsqueeze(1)
        std = self.scales[:, 1].unsqueeze(1)
        return SeriesBatch(
            self.sequences * std + m,
            self.lengths,
            self.split_sections,
        )

    def one_per_split(self) -> SeriesBatch:
        """
        Choose the first element of every split section and return the
        resulting series batch.

        This method should only be used on query (past and future) series
        batches.
        """
        splits = torch.split(self.sequences, self.split_sections.tolist())
        lengths = torch.split(self.lengths, self.split_sections.tolist())
        return SeriesBatch(
            sequences=torch.cat(tuple(split[0:1] for split in splits)),
            lengths=torch.cat(tuple(l[0:1] for l in lengths)),
            split_sections=torch.ones(len(splits), dtype=int),
        )

    def first_n(self, n: int) -> SeriesBatch:
        """
        Choose the first n splits of the sequences as defined by
        ``split_sections``.

        Self.sequences has thus sum(split_section[i], i=0, ..., n-1) elements.
        """
        splits = torch.split(self.sequences, self.split_sections.tolist())
        lengths = torch.split(self.lengths, self.split_sections.tolist())
        return SeriesBatch(
            sequences=torch.cat(splits[:n]),
            lengths=torch.cat(lengths[:n]),
            split_sections=self.split_sections[:n],
        )

    def __getitem__(self, index: int) -> SeriesBatch:
        split = torch.split(self.sequences, self.split_sections.tolist())[
            index
        ]
        length = torch.split(self.lengths, self.split_sections.tolist())[index]
        return SeriesBatch(
            sequences=split,
            lengths=length,
            split_sections=self.split_sections[index : index + 1],
        )


@dataclass
class TripletBatch:
    """
    A triplet batch, composed of a batch of support sets, query contexts and
    prediction horizons.

    Compared to a simple triplet, it also manages the lengths of all samples.
    """

    support_set: SeriesBatch
    query_past: SeriesBatch
    query_future: SeriesBatch

    @classmethod
    def collate(cls, triplets: List[Triplet]) -> TripletBatch:
        """
        Combines a list of triplets into a batched triplet to pass to a
        network.
        """
        s, p, f = zip(*triplets)
        return TripletBatch(
            SeriesBatch.from_lists(list(s)),
            SeriesBatch.from_lists(list(p)),
            SeriesBatch.from_lists(list(f), squeeze=True),
        )

    def to(self, device) -> TripletBatch:
        return TripletBatch(
            self.support_set.to(device),
            self.query_past.to(device),
            self.query_future.to(device),
        )

    def pin_memory(self):
        self.support_set = self.support_set.pin_memory()
        self.query_past = self.query_past.pin_memory()
        self.query_future = self.query_future.pin_memory()
        return self

    def reduce_to_unique_query(self) -> TripletBatch:
        """
        Selects the first query of every group of queries that use the same
        support set (see split_sections of queries) and returns the resulting
        triplet batch.

        Query past and future are reduced, support sets are not touched.
        """
        return TripletBatch(
            support_set=self.support_set,
            query_past=self.query_past.one_per_split(),
            query_future=self.query_future.one_per_split(),
        )

    def first_n(self, n: int) -> TripletBatch:
        """
        Choose the first n splits of each series batch and return the resulting
        triplet batch.
        """
        return TripletBatch(
            support_set=self.support_set.first_n(n),
            query_past=self.query_past.first_n(n),
            query_future=self.query_future.first_n(n),
        )

    def __getitem__(self, index: int) -> TripletBatch:
        return TripletBatch(
            support_set=self.support_set[index],
            query_past=self.query_past[index],
            query_future=self.query_future[index],
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
import warnings
import json
from typing import Dict, Union, Optional
from collections import OrderedDict

from .meters import Meter


class MeterDict(object):
    def __init__(
        self,
        meters: Optional[Dict[str, Meter]] = None,
        meterdicts: Optional[Dict[str, MeterDict]] = None,
    ) -> None:
        self._meters = OrderedDict()
        self._meterdicts = OrderedDict()
        if meters is not None:
            for name, meter in meters.items():
                self.register_meter(name, meter)
        if meterdicts is not None:
            for name, meterdict in meterdicts.items():
                self.register_meterdict(name, meterdict)

    def register_meter(self, name: str, meter: Meter) -> None:
        if "_meters" not in self.__dict__:
            raise AttributeError(
                "cannot assign meter before MeterDict.__init__() call"
            )
        elif "/" in name:
            raise KeyError("meter name cannot contain '/'")
        elif name == "":
            raise KeyError("meter name cannot be empty")
        elif hasattr(self, name) and name not in self._meters:
            raise KeyError(f"attribute {name} already exists")
        self._meters[name] = meter

    def register_meterdict(self, name: str, meterdict: MeterDict) -> None:
        if "_meterdicts" not in self.__dict__:
            raise AttributeError(
                "cannot assign meterdict before MeterDict.__init__() call"
            )
        elif "/" in name:
            raise KeyError("meterdict name cannot contain '/'")
        elif name == "":
            raise KeyError("meterdict name cannot be empty")
        elif hasattr(self, name) and name not in self._meters:
            raise KeyError(f"attribute {name} already exists")
        self._meterdicts[name] = meterdict

    def __setattr__(self, name: str, value):
        if isinstance(value, Meter):
            if name in self.__dict__:
                del self.__dict__[name]
            if name in self._meterdicts:
                del self._meterdicts[name]
            self.register_meter(name, value)
        elif isinstance(value, MeterDict):
            if name in self.__dict__:
                del self.__dict__[name]
            if name in self._meters:
                del self._meters[name]
            self.register_meterdict(name, value)
        else:
            super(MeterDict, self).__setattr__(name, value)

    def __getattr__(self, name: str):
        if "_meters" in self.__dict__:
            _meters = self.__dict__["_meters"]
            if name in _meters:
                return _meters[name]
        if "_meterdicts" in self.__dict__:
            _meterdicts = self.__dict__["_meterdicts"]
            if name in _meterdicts:
                return _meterdicts[name]
        raise AttributeError(f"{type(self).__name__} has no attribute {name}")

    def __delattr__(self, name: str):
        if name in self._meters:
            del self._meters[name]
        elif name in self._meterdicts:
            del self._meterdicts[name]
        else:
            object.__delattr__(name)

    def __setitem__(self, name: str, value):
        names = name.split("/", 1)
        if len(names) > 1:
            member = getattr(self, names[0])
            member[names[1]] = value
        else:
            setattr(self, names[0], value)

    def __getitem__(self, name: str):
        names = name.split("/", 1)
        item = getattr(self, names[0])
        if len(names) > 1:
            return item[names[1]]
        else:
            return item

    def __contains__(self, name: str):
        names = name.split("/", 1)
        if hasattr(self, names[0]):
            if len(names) > 1:
                return names[1] in getattr(self, names[0])
            else:
                return True
        else:
            return False

    def get(self, name: str, default=None):
        if name in self:
            return self[name]
        else:
            return default

    def _named_dicts(self, memo: Optional[set] = None, prefix: str = ""):
        if memo is None:
            memo = set()
        if self not in memo:
            memo.add(self)
            yield prefix, self
            if len(prefix) > 0:
                prefix += "/"
            for name, meterdict in self._meterdicts.items():
                for m in meterdict._named_dicts(memo, prefix + name):
                    yield m

    def _named_meters(self, prefix: str = ""):
        memo = set()
        for name_prefix, meterdict in self._named_dicts(prefix=prefix):
            for name, meter in meterdict._meters.items():
                if meter in memo:
                    continue
                memo.add(meter)
                name = (
                    name_prefix + ("/" if len(name_prefix) > 0 else "") + name
                )
                yield name, meter

    def restart(self):
        for _, meter in self._named_meters():
            meter.restart()

    @property
    def value(self) -> Dict:
        return {name: meter.value for name, meter in self._named_meters()}

    @property
    def best(self) -> Dict:
        return {name: meter.best for name, meter in self._named_meters()}

    def state_dict(self) -> Dict:
        return {
            name: meter.state_dict() for name, meter in self._named_meters()
        }

    def load_state_dict(self, state_dict: Dict) -> None:
        missing_keys = []
        for name, meter in self._named_meters():
            if name in state_dict:
                meter_state = state_dict.pop(name)
                meter.load_state_dict(meter_state)
            else:
                missing_keys.append(name)
        unexpected_keys = list(state_dict.keys())
        if len(missing_keys) > 0:
            warnings.warn(f"missing keys in state_dict: {missing_keys}")
        if len(unexpected_keys) > 0:
            warnings.warn(f"unexpected keys in state_dict: {unexpected_keys}")

    def __repr__(self) -> str:
        def _add_spaces(str_, n_spaces=4):
            return "\n".join(
                [(" " * n_spaces) + line for line in str_.split("\n")]
            )

        main_str = "\n".join(
            [f"{name}: {repr(meter)}" for name, meter in self._meters.items()]
        )
        child_str = "\n".join(
            [
                f"{name}:\n{_add_spaces(repr(meterdict))}"
                for name, meterdict in self._meterdicts.items()
            ]
        )
        if child_str:
            main_str = "\n".join([main_str, child_str])
        return main_str



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import json
from itertools import cycle
from pathlib import Path
from typing import List, Optional, Any, Dict
from src.gluonts.dataset.field_names import FieldName

from .filters import Filter


Item = Dict[str, Any]


class Transform(ABC):
    """
    A transform enables transforming the set of time series contained in a
    dataset.
    """

    @abstractmethod
    def __call__(self, items: List[Item]) -> List[Item]:
        """
        Transforms the given items and returns the transformed ones.

        Args:
            items:  The items to transform.

        Returns:
            The transformed items.
        """


class ItemIDTransform(Transform):
    """
    Adds an id to the time series.

    For rolling test sets the time series that are an extension of each other
    have the same id.
    """

    def __init__(self, required_length: int = 0):
        self.required_length = required_length

    def __call__(self, items: List[Item]) -> List[Item]:
        assert (not self.required_length) or (
            not len(items) % self.required_length
        ), "Either no length specified or number of times series is a multiple of specified length."
        self.required_length = self.required_length or len(items)
        it_ids = cycle(range(self.required_length))
        for item in items:
            item[FieldName.ITEM_ID] = next(it_ids)
        return items


def read_transform_write(
    file: Path,
    filters: Optional[List[Filter]] = None,
    source: Optional[Path] = None,
) -> None:
    """
    Reads the dataset from the provided path, applies the given transform and
    writes it back to the same file.

    Args:
        file: The path from where to read the data.
        filters: An optional list of filters to apply to modify the items in the datset.
        source: An optional source to read from. Defaults to the path to write to.
    """
    # Read
    data = []
    with (source or file).open("r") as f:
        for line in f:
            data.append(json.loads(line))

    # Filter
    for f in filters or []:
        data = f(data)

    # Write
    file.parent.mkdir(parents=True, exist_ok=True)
    with file.open("w") as f:
        content = "\n".join([json.dumps(d) for d in data])
        f.write(content + "\n")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import List
import numpy as np
import numpy.typing as npt
import pandas as pd
from src.gluonts.model.forecast import QuantileForecast
from src.gluonts.time_feature import get_seasonality


@dataclass
class QuantileForecasts:
    """
    A type-safe wrapper for a list of quantile forecasts, stored as NumPy
    arrays.
    """

    values: npt.NDArray[
        np.float32
    ]  # [num_time_series, num_quantiles, prediction_length]
    start_dates: npt.NDArray[np.float32]
    item_ids: npt.NDArray[np.float32]
    freq: pd.DateOffset
    quantiles: list[str]

    @property
    def prediction_length(self) -> int:
        """
        Returns the prediction length of the quantile forecasts.
        """
        return self.values.shape[-1]

    @property
    def seasonality(self) -> int:
        """
        Returns the seasonality of the forecasts (i.e. how many steps to go
        back to arrive at the value of the previous period).
        """
        return get_seasonality(self.freq.freqstr)  # type: ignore

    # ---------------------------------------------------------------------------------------------
    # DATA ACCESS

    def get(self, index: int) -> QuantileForecast:
        """
        Returns the quantile forecast at the specified index.

        This method should typically only be used for visualizing single
        forecasts.
        """
        return QuantileForecast(
            forecast_arrays=self.values[index],
            start_date=pd.Timestamp(self.start_dates[index], freq=self.freq),
            freq=self.freq.freqstr,  # type: ignore
            item_id=self.item_ids[index],
            forecast_keys=self.quantiles,
        )

    @property
    def median(self) -> npt.NDArray[np.float32]:
        """
        Returns the median forecasts for all time series.

        NumPy array of shape [N, T] (N: number of forecasts, T: forecast
        horizon).
        """
        i = self.quantiles.index("0.5")
        return self.values[:, i]

    def __len__(self) -> int:
        return self.values.shape[0]

    def __getitem__(self, index: npt.ArrayLike) -> QuantileForecasts:  # type: ignore
        return QuantileForecasts(
            values=self.values[index],
            start_dates=self.start_dates[index],
            item_ids=self.item_ids[index],
            freq=self.freq,
            quantiles=self.quantiles,
        )

    # ---------------------------------------------------------------------------------------------
    # STORAGE

    @classmethod
    def load(cls, path: Path) -> QuantileForecasts:
        """
        Loads the quantile forecasts from the specified path.

        Args:
            path: The path from where to load the forecasts.
        """
        try:
            with (path / "values.npy").open("rb") as f:
                values = np.load(f)
        except FileNotFoundError:
            # This is a bug in the publicly available evaluations for MQ-RNN.
            # Should not be called otherwise.
            with (path / "values.npz").open("rb") as f:
                values = np.load(f)
        with (path / "metadata.npz").open("rb") as f:
            metadata = np.load(f, allow_pickle=True).item()
        return QuantileForecasts(
            values=values,
            start_dates=metadata["start_dates"],
            item_ids=metadata["item_ids"],
            freq=metadata["freq"],
            quantiles=metadata["quantiles"],
        )

    def save(self, path: Path) -> None:
        """
        Saves the forecasts to the specified path.

        Args:
            path: The path of the file where to save the forecasts to.
        """
        assert path.is_dir()

        if not path.exists():
            path.mkdir()
        with (path / "values.npy").open("wb+") as f:
            np.save(f, self.values)
        with (path / "metadata.npz").open("wb+") as f:
            np.save(
                f,
                dict(
                    start_dates=self.start_dates,
                    item_ids=self.item_ids,
                    freq=self.freq,
                    quantiles=self.quantiles,
                ),
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from pathlib import Path
from typing import Optional
import numpy as np
import pandas as pd

from src.gluonts import json
from src.gluonts.dataset.repository._m3 import check_dataset
from src.gluonts.dataset.repository._m3 import M3Setting as M1Setting
from src.gluonts.dataset.repository._util import metadata, save_to_file, to_dict

from meta.datasets.gluonts import GluonTSDataModule
from meta.datasets.registry import register_data_module


@register_data_module
class M1DataModule(GluonTSDataModule):
    """
    A data module which provides a frequency-category split of the M1 dataset
    as a standalone dataset.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.freq, self.category = self.dataset_name.split("_")[1:]

    def _materialize(self, directory: Path) -> None:
        generate_m1_dataset(
            dataset_path=directory / self.dataset_name,
            m1_freq=self.freq,
            category=self.category,
        )

    @classmethod
    def name(cls) -> str:
        return "dm_m1"


def generate_m1_dataset(
    dataset_path: Path,
    m1_freq: str,
    category: str,
    prediction_length: Optional[int] = None,
):
    from src.gluonts.dataset.repository import default_dataset_path

    m1_xls_path = default_dataset_path / "MC1001.xls"
    if not os.path.exists(m1_xls_path):
        raise RuntimeError(
            f"The M1 data is available at https://forecasters.org/resources/time-series-data/m-competition/"
            f"Please download the file and copy the files to this location: {m1_xls_path}"
        )

    subsets = {
        "yearly": M1Setting("M1Year", 6, "Y"),
        "quarterly": M1Setting("M1Quart", 8, "Q"),
        "monthly": M1Setting("M1Month", 18, "M"),
    }
    assert (
        m1_freq.lower() in subsets
    ), f"invalid m1_freq='{m1_freq}'. Allowed values: {subsets.keys()}"

    subset = subsets[m1_freq.lower()]
    df = pd.read_excel(m1_xls_path, sheet_name="MC1001")

    def truncate_trailing_nan(v: np.ndarray):
        last_finite_index = np.where(np.isfinite(v))[0][-1]
        return v[: last_finite_index + 1]

    train_data = []
    test_data = []

    def normalize_str(c: str):
        return c.strip().lower()

    # select category
    df["Category"] = df["Category"].apply(normalize_str)
    categories = list(df["Category"].unique())
    assert category in categories, f"category must be one of {categories}"
    df = df.loc[df["Category"] == category]

    # select frequency
    df["Type"] = df["Type"].apply(normalize_str)
    frequencies = list(df["Type"].unique())
    assert m1_freq in frequencies, f"frequency must be one of {frequencies}"
    df = df.loc[df["Type"] == m1_freq]

    i = 0
    for _, row in df.iterrows():
        vals = row.values
        series, n, seasonality, nf, freq, starting_date, category = vals[:7]
        target = np.asarray(vals[7:], dtype=np.float64)
        target = truncate_trailing_nan(target)
        assert len(target) == n + nf
        assert nf == subset.prediction_length

        starting_date = str(starting_date)
        # correct some date errors
        if starting_date.startswith("19.."):
            # use mock date for invalid dates
            starting_date = "1900"
        if starting_date.strip().endswith("JUI"):
            # set to June (?)
            starting_date = starting_date[:-3] + "JUN"
        if starting_date.strip().endswith("JUJUN"):
            starting_date = starting_date[:-5] + "JUN"
        if starting_date.strip().endswith("AVR"):
            starting_date = starting_date[:-3] + "APR"
        if starting_date.strip().endswith("AVJAN"):
            starting_date = starting_date[:-5] + "JAN"
        if starting_date.strip().endswith("DE"):
            starting_date = starting_date[:-3] + "DEC"
        if starting_date.strip().endswith("DEDEC"):
            starting_date = starting_date[:-5] + "DEC"
        if starting_date.strip().endswith("AVAPR"):
            starting_date = starting_date[:-5] + "APR"
        s = pd.Timestamp(starting_date, freq=subset.freq)
        s = s.freq.rollforward(s)
        start = str(s).split(" ")[0]

        d_train = to_dict(
            target_values=target[: -subset.prediction_length],
            start=start,
            item_id=series,
        )
        train_data.append(d_train)

        d_test = to_dict(
            target_values=target,
            start=start,
            item_id=series,
        )
        test_data.append(d_test)
        i += 1

    os.makedirs(dataset_path, exist_ok=True)
    with open(dataset_path / "metadata.json", "w") as f:
        f.write(
            json.dumps(
                metadata(
                    cardinality=len(train_data),
                    freq=subset.freq,
                    prediction_length=prediction_length
                    or subset.prediction_length,
                )
            )
        )

    train_file = dataset_path / "train" / "data.json"
    test_file = dataset_path / "test" / "data.json"

    save_to_file(train_file, train_data)
    save_to_file(test_file, test_data)

    check_dataset(dataset_path, len(df), subset.sheet_name)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional
import itertools

from torch.utils.data import IterableDataset

from src.gluonts.dataset.common import Dataset
from src.gluonts.transform import Transformation, TransformedDataset
from src.gluonts.itertools import Cyclic, PseudoShuffled, Cached


class TransformedIterableDataset(IterableDataset):
    def __init__(
        self,
        dataset: Dataset,
        transform: Transformation,
        is_train: bool = True,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
    ):
        super().__init__()
        self.shuffle_buffer_length = shuffle_buffer_length

        self.transformed_dataset = TransformedDataset(
            Cyclic(dataset) if not cache_data else Cached(Cyclic(dataset)),
            transform,
            is_train=is_train,
        )

    def __iter__(self):
        if self.shuffle_buffer_length is None:
            return iter(self.transformed_dataset)
        else:
            shuffled = PseudoShuffled(
                self.transformed_dataset,
                shuffle_buffer_length=self.shuffle_buffer_length,
            )
            return iter(shuffled)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Callable, Optional, List, Tuple, Dict, Union
from functools import partial
from itertools import chain
from pathlib import Path

from torch import Tensor, BoolTensor
from torch.utils.data import Dataset
from torch.optim import AdamW
import numpy as np

from tslib.dataset import MetaDataset
from tslib.engine import Trainer, Evaluator, HyperOptManager
from tslib.engine.distributed import (
    reduce_value,
    is_main_process,
)
from tslib.engine.callback import *
from tslib.metrics import (
    NumericalAverageMeter,
    BatchAverageMeter,
    MeanDeviationMeter,
    MeterDict,
)
from ..estimator import (
    DomAdaptEstimator,
    AdversarialDomAdaptEstimator,
)


class DomAdaptTrainer(Trainer):
    def _train(self, *data):
        loss = self.model(*data)
        loss.mean().backward()
        loss = reduce_value(loss, dst=None)
        self.metrics.train.loss.update(loss)
        self.metrics.train.src.bc_loss.update(self.model.src.bc_loss)
        self.metrics.train.src.fc_loss.update(self.model.src.fc_loss)
        self.metrics.train.tgt.bc_loss.update(self.model.tgt.bc_loss)
        self.metrics.train.tgt.fc_loss.update(self.model.tgt.fc_loss)

    def _validate(self, *data):
        _ = self.model(*data)
        self.metrics.valid.src.bc_loss.update(self.model.src.bc_loss)
        self.metrics.valid.src.fc_loss.update(self.model.src.fc_loss)
        self.metrics.valid.src.fc_ND.update(
            self.model.src.fc_loss, self.model.src.denominator
        )
        self.metrics.valid.tgt.bc_loss.update(self.model.tgt.bc_loss)
        self.metrics.valid.tgt.fc_loss.update(self.model.tgt.fc_loss)
        self.metrics.valid.tgt.fc_ND.update(
            self.model.tgt.fc_loss, self.model.tgt.denominator
        )

    @classmethod
    def from_configs(
        cls,
        dataset: MetaDataset,
        model: DomAdaptEstimator,
        log_dir: Union[Path, str],
        cuda_device: int,
        n_epochs: int,
        nb_epoch: Optional[int] = None,
        batch_size: Union[int, Tuple[int, int]] = 100,
        max_grad_norm: float = 1e0,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        weight_decay: float = 0.0,
        amsgrad: bool = False,
        debug: bool = False,
        **kwargs,
    ):
        optimizer = partial(
            AdamW,
            lr=lr,
            betas=betas,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )

        meters = MeterDict(
            meterdicts={
                "train": MeterDict(
                    meters=dict(loss=BatchAverageMeter()),
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                            )
                        ),
                    ),
                ),
                "valid": MeterDict(
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                    ),
                ),
                "test": MeterDict(
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                    ),
                ),
            }
        )

        callbacks = [
            Optimization(max_grad_norm=max_grad_norm),
            ProgressBar(
                batch_monitors=[
                    "train/loss",
                    "train/src/fc_loss",
                    "train/tgt/fc_loss",
                ],
                epoch_monitors=[
                    "train/src/fc_loss",
                    "train/tgt/fc_loss",
                    "valid/src/fc_loss",
                    "valid/tgt/fc_loss",
                    "valid/src/fc_ND",
                    "valid/tgt/fc_ND",
                ],
            ),
            Checkpoint(monitor="valid/tgt/fc_loss"),
            Tensorboard(
                batch_monitors=[
                    "train/loss",
                    "train/src/bc_loss",
                    "train/src/fc_loss",
                    "train/tgt/bc_loss",
                    "train/tgt/fc_loss",
                ],
                epoch_monitors=[
                    "valid/src/bc_loss",
                    "valid/src/fc_loss",
                    "valid/src/fc_ND",
                    "valid/tgt/bc_loss",
                    "valid/tgt/fc_loss",
                    "valid/tgt/fc_ND",
                ],
            ),
        ]

        return cls(
            dataset,
            model,
            optimizer,
            None,
            metrics=meters,
            callbacks=callbacks,
            log_dir=log_dir,
            n_epochs=n_epochs,
            nb_epoch=nb_epoch,
            batch_size=batch_size,
            cuda_device=cuda_device,
            debug=debug,
        )


class DomAdaptEvaluator(Evaluator):
    def _test(self, *data):
        _ = self.model(*data)
        self.metrics.test.src.bc_loss.update(self.model.src.bc_loss)
        self.metrics.test.src.fc_loss.update(self.model.src.fc_loss)
        self.metrics.test.src.fc_ND.update(
            self.model.src.fc_loss, self.model.src.denominator
        )
        self.metrics.test.tgt.bc_loss.update(self.model.tgt.bc_loss)
        self.metrics.test.tgt.fc_loss.update(self.model.tgt.fc_loss)
        self.metrics.test.tgt.fc_ND.update(
            self.model.tgt.fc_loss, self.model.tgt.denominator
        )

    def predict(self, dataset: Optional[Dataset] = None) -> np.ndarray:
        if dataset is None:
            loader = self.test_loader
            data_size = self.dataset.test_size
        else:
            loader = MetaDataset._data_loader(
                dataset=dataset,
                batch_size=self.batch_size,
                shuffle=False,
                cuda_device=self.cuda_device,
                is_training=False,
                n_workers=self.n_loader_workers,
                n_batches=None,
            )
            data_size = len(dataset)
        device = "cpu" if self.cuda_device < 0 else f"cuda:{self.cuda_device}"
        prog_bar = tqdm(
            desc=f"({device}) test",
            total=int(math.ceil(data_size / self.batch_size)),
            unit="batch",
        )
        _ = self.model.eval()
        src_preds, tgt_preds = [], []
        for batch, data in enumerate(loader):
            with pt.no_grad():
                src_result, tgt_result = self._predict(*data)
            src_preds.append(src_result)
            tgt_preds.append(tgt_result)
            prog_bar.update(1)
            if self.debug and batch > 0:
                break
        prog_bar.close()
        src_preds = np.concatenate(src_preds, axis=0)
        tgt_preds = np.concatenate(tgt_preds, axis=0)
        return src_preds, tgt_preds

    def _predict(self, *data) -> np.ndarray:
        _ = self.model(*data)
        src_truth = data[0].cpu().numpy()
        tgt_truth = data[1].cpu().numpy()
        src_preds = self.model.src.forecast[..., 0].cpu().numpy()
        tgt_preds = self.model.tgt.forecast[..., 0].cpu().numpy()
        src_dummy = (
            np.zeros(src_truth.shape[:-1] + src_preds.shape[-1:]) + np.nan
        )
        src_preds = np.concatenate(
            [src_dummy[:, : -src_preds.shape[1]], src_preds], axis=1
        )
        src_curve = np.concatenate([src_truth, src_preds], axis=-1)
        tgt_dummy = (
            np.zeros(tgt_truth.shape[:-1] + tgt_preds.shape[-1:]) + np.nan
        )
        tgt_preds = np.concatenate(
            [tgt_dummy[:, : -tgt_preds.shape[1]], tgt_preds], axis=1
        )
        tgt_curve = np.concatenate([tgt_truth, tgt_preds], axis=-1)
        return src_curve, tgt_curve

    @classmethod
    def from_trainer(
        cls,
        trainer: DomAdaptTrainer,
        model_tag: str = "best",
    ):
        return super(DomAdaptEvaluator, cls).from_trainer(trainer, model_tag)


class AdversarialDomAdaptTrainer(DomAdaptTrainer):
    def __init__(
        self,
        dataset: MetaDataset,
        model: AdversarialDomAdaptEstimator,
        optimizer: partial,
        scheduler: Optional[partial],
        **kwargs,
    ) -> None:
        super(AdversarialDomAdaptTrainer, self).__init__(
            dataset,
            model,
            optimizer,
            scheduler,
            **kwargs,
        )
        del self.optimizer
        self.optimizer = optimizer(
            [
                {
                    "params": chain(
                        self.model.src.generative_parameters(),
                        self.model.tgt.generative_parameters(),
                    )
                },
                {
                    "params": chain(
                        self.model.src.discriminative_parameters(),
                        self.model.tgt.discriminative_parameters(),
                    )
                },
            ]
        )

    def _train(self, *data):
        self.model.generative()
        gen_loss = self.model(*data)
        gen_loss.mean().backward()
        gen_loss = reduce_value(gen_loss, dst=None)
        self.metrics.train.loss.update(gen_loss)
        self.metrics.train.src.bc_loss.update(self.model.src.bc_loss)
        self.metrics.train.src.fc_loss.update(self.model.src.fc_loss)
        self.metrics.train.tgt.bc_loss.update(self.model.tgt.bc_loss)
        self.metrics.train.tgt.fc_loss.update(self.model.tgt.fc_loss)

        self.model.discriminative()
        disc_loss = self.model(*data)
        disc_loss.mean().backward()
        disc_loss = reduce_value(disc_loss, dst=None)
        self.metrics.train.disc.update(disc_loss)

    def _validate(self, *data):
        self.model.generative()
        _ = self.model(*data)
        self.metrics.valid.src.bc_loss.update(self.model.src.bc_loss)
        self.metrics.valid.src.fc_loss.update(self.model.src.fc_loss)
        self.metrics.valid.tgt.bc_loss.update(self.model.tgt.bc_loss)
        self.metrics.valid.tgt.fc_loss.update(self.model.tgt.fc_loss)

    @classmethod
    def from_configs(
        cls,
        dataset: MetaDataset,
        model: AdversarialDomAdaptEstimator,
        log_dir: Union[Path, str],
        cuda_device: int,
        n_epochs: int,
        nb_epoch: Optional[int] = None,
        batch_size: Union[int, Tuple[int, int]] = 100,
        max_grad_norm: float = 1e0,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        weight_decay: float = 0.0,
        amsgrad: bool = False,
        debug: bool = False,
        **kwargs,
    ):
        optimizer = partial(
            AdamW,
            lr=lr,
            betas=betas,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )

        meters = MeterDict(
            meterdicts={
                "train": MeterDict(
                    meters=dict(
                        loss=BatchAverageMeter(),
                        disc=BatchAverageMeter(),
                        grad_norm=NumericalAverageMeter(),
                    ),
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                            )
                        ),
                    ),
                ),
                "valid": MeterDict(
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                    ),
                ),
                "test": MeterDict(
                    meterdicts=dict(
                        src=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                        tgt=MeterDict(
                            meters=dict(
                                bc_loss=BatchAverageMeter(),
                                fc_loss=BatchAverageMeter(),
                                fc_ND=MeanDeviationMeter(),
                            )
                        ),
                    ),
                ),
            }
        )

        callbacks = [
            Optimization(max_grad_norm=max_grad_norm),
            ProgressBar(
                batch_monitors=[
                    "train/loss",
                    "train/disc",
                    "train/src/fc_loss",
                    "train/tgt/fc_loss",
                ],
                epoch_monitors=[
                    "train/src/fc_loss",
                    "train/tgt/fc_loss",
                    "valid/src/fc_loss",
                    "valid/tgt/fc_loss",
                    "valid/src/fc_ND",
                    "valid/tgt/fc_ND",
                ],
            ),
            Checkpoint(monitor="valid/tgt/fc_loss"),
            Tensorboard(
                batch_monitors=[
                    "train/loss",
                    "train/disc",
                    "train/src/bc_loss",
                    "train/src/fc_loss",
                    "train/tgt/bc_loss",
                    "train/tgt/fc_loss",
                ],
                epoch_monitors=[
                    "valid/src/bc_loss",
                    "valid/src/fc_loss",
                    "valid/src/fc_ND",
                    "valid/tgt/bc_loss",
                    "valid/tgt/fc_loss",
                    "valid/tgt/fc_ND",
                ],
            ),
        ]

        return cls(
            dataset,
            model,
            optimizer,
            None,
            metrics=meters,
            callbacks=callbacks,
            log_dir=log_dir,
            n_epochs=n_epochs,
            nb_epoch=nb_epoch,
            batch_size=batch_size,
            cuda_device=cuda_device,
            debug=debug,
        )


class AdversarialDomAdaptEvaluator(DomAdaptEvaluator):
    @classmethod
    def from_trainer(
        cls,
        trainer: AdversarialDomAdaptTrainer,
        model_tag: str = "best",
    ):
        return super(AdversarialDomAdaptEvaluator, cls).from_trainer(
            trainer, model_tag
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pts.dataset.repository.datasets import get_dataset, dataset_recipes
from pts.dataset.utils import to_pandas
import numpy as np
import pandas as pd
from pts.dataset import ListDataset
import torch
from pts.model.simple_feedforward import SimpleFeedForwardEstimator
import pickle
import json
import random
import time
from matplotlib import pyplot as plt
from pykeops.torch import LazyTensor


def KMeans(x, K=10, Niter=10000, verbose=True):
    """
    Implements Lloyd's algorithm for the Euclidean metric.
    """
    use_cuda = torch.cuda.is_available()
    dtype = torch.float32 if use_cuda else torch.float64

    start = time.time()
    N, D = x.shape  # Number of samples, dimension of the ambient space
    c = x[:K, :].clone()  # Simplistic initialization for the centroids

    x_i = LazyTensor(x.view(N, 1, D))  # (N, 1, D) samples
    c_j = LazyTensor(c.view(1, K, D))  # (1, K, D) centroids

    # K-means loop:
    # - x  is the (N, D) point cloud,
    # - cl is the (N,) vector of class labels
    # - c  is the (K, D) cloud of cluster centroids
    for i in range(Niter):
        # E step: assign points to the closest cluster -------------------------
        D_ij = ((x_i - c_j) ** 2).sum(-1)  # (N, K) symbolic squared distances
        # print(D_ij.argmin(dim=1))
        # import pdb;pdb.set_trace()
        cl = D_ij.argmin(dim=1).long().view(-1)  # Points -> Nearest cluster

        # M step: update the centroids to the normalized cluster average: ------
        # Compute the sum of points per cluster:
        c.zero_()
        c.scatter_add_(0, cl[:, None].repeat(1, D), x)

        # Divide by the number of points per cluster:
        Ncl = torch.bincount(cl, minlength=K).type_as(c).view(K, 1)
        c /= Ncl  # in-place division to compute the average

    if (
        verbose
    ):  # Fancy display -----------------------------------------------
        if use_cuda:
            torch.cuda.synchronize()
        end = time.time()
        print(
            f"K-means for the Euclidean metric with {N:,} points in dimension"
            f" {D:,}, K = {K:,}:"
        )
        print(
            "Timing for {} iterations: {:.5f}s = {} x {:.5f}s\n".format(
                Niter, end - start, Niter, (end - start) / Niter
            )
        )

    return cl, c


def _get_pre_features(x):
    df = pd.DataFrame(x.numpy())
    df = df.rename(
        columns={
            0: "mean",
            1: "var",
            2: "week_id",
            3: "month_id",
        }
    )
    df["mean"] = (df["mean"] - df["mean"].min()) / (
        df["mean"].max() - df["mean"].min()
    )
    df["var"] = (df["var"] - df["var"].min()) / (
        df["var"].max() - df["var"].min()
    )
    df = pd.get_dummies(df, columns=["week_id", "month_id"], dummy_na=True)
    return torch.from_numpy(df.to_numpy())


def KMeans_inside_dataset(
    num_ts_=1,
    num_groups=16,
    context_length=72,
    prediction_length=24,
    file_name="default",
):
    dataset = get_dataset("traffic")
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = dict()
    it = iter(dataset.train)
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    num_ts = num_ts_
    len_sample = context_length + prediction_length
    index = 0
    feature = torch.Tensor([])
    for i in range(num_ts):
        train_entry = next(it)
        target = train_entry["target"]

        for ts_sample_start in range(
            0, len(target) - len_sample, prediction_length
        ):
            ts_slice = target[ts_sample_start : ts_sample_start + len_sample]
            feature = torch.cat(
                (
                    feature,
                    torch.Tensor(
                        [
                            ts_slice.mean(),
                            ts_slice.var(),
                            index % 7,
                            index // 90,
                        ]
                    ),
                )
            )
            index += 1
    feature = feature.reshape(index, 4)
    feature = _get_pre_features(feature).contiguous()
    # print(feature)
    # import pdb;pdb.set_trace()
    cl, c = KMeans(feature, num_groups)
    it = iter(dataset.train)
    sample_id = 0
    for i in range(num_ts):
        train_entry = next(it)
        target = train_entry["target"]
        unsplit_start = train_entry["start"]
        for ts_sample_start in range(
            0, len(target) - len_sample, prediction_length
        ):
            ts_slice = target[ts_sample_start : ts_sample_start + len_sample]
            gid = cl[sample_id]
            dataset_group[gid].append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            whole_data.append(
                {
                    "target": ts_slice,
                    "start": unsplit_start,
                    "feat_static_cat": train_entry["feat_static_cat"],
                }
            )
            unsplit_start += pd.Timedelta(hours=prediction_length)
            sample_id += 1
    print(len(whole_data))
    ret["group_ratio"] = [len(i) / len(whole_data) for i in dataset_group]
    print(ret["group_ratio"])
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq=dataset.metadata.freq)
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq=dataset.metadata.freq))
    ret["group_data"] = group_data_list
    print("write whole data")
    with open("synthetic_" + file_name + "_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)
    print("write group data")
    with open("synthetic_" + file_name + "_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True


def KMeans_m5_dataset(
    num_ts_=1,
    num_groups=16,
    context_length=72,
    prediction_length=24,
    file_name="default",
):
    df = pd.read_csv("sales_train_evaluation.csv")
    dataset_group = [[] for i in range(num_groups)]
    whole_data = []
    ret = dict()
    # num_ts = int(dataset.metadata.feat_static_cat[0].cardinality)
    num_ts = num_ts_
    len_sample = context_length + prediction_length
    # compute mean and variance
    df = df.iloc[:num_ts, :]
    df["mean"] = df.iloc[:, 1947 - len_sample : 1947].mean(axis=1)
    df["var"] = df.iloc[:, 1947 - len_sample : 1947].var(axis=1)
    df["mean"] = (df["mean"] - df["mean"].min()) / (
        df["mean"].max() - df["mean"].min()
    )
    df["var"] = (df["var"] - df["var"].min()) / (
        df["var"].max() - df["var"].min()
    )
    df_feature = df.iloc[:, 2:6]
    df_feature = pd.get_dummies(df_feature, dummy_na=True)
    df_feature = pd.concat([df_feature, df.iloc[:, -2]], axis=1)
    feature = torch.from_numpy(df_feature.to_numpy()).contiguous()
    cl, c = KMeans(feature, num_groups)
    # print(cl)
    # import pdb;pdb.set_trace()
    sample_id = 0
    for i in range(num_ts):
        ts_slice = df.iloc[i : i + 1, 1947 - len_sample : 1947]
        ts_slice = torch.from_numpy(ts_slice.to_numpy())[0]
        # print(ts_slice)
        # import pdb;pdb.set_trace()
        gid = cl[sample_id]
        unsplit_start = pd.Timestamp("1990-01-01")
        dataset_group[gid].append(
            {
                "target": ts_slice,
                "start": unsplit_start,
            }
        )  # , 'feat_static_cat': train_entry['feat_static_cat']}
        whole_data.append(
            {
                "target": ts_slice,
                "start": unsplit_start,
            }
        )  # , 'feat_static_cat': train_entry['feat_static_cat']}
        sample_id += 1
    print(len(whole_data))
    ret["group_ratio"] = [len(i) / len(whole_data) for i in dataset_group]
    print(ret["group_ratio"])
    random.shuffle(whole_data)
    ret["whole_data"] = ListDataset(whole_data, freq="1H")
    group_data_list = []
    for group in dataset_group:
        random.shuffle(group)
        group_data_list.append(ListDataset(group, freq="1H"))
    ret["group_data"] = group_data_list
    print("write whole data")
    with open("synthetic_" + file_name + "_whole_data.csv", "wb") as output:
        pickle.dump(ret["whole_data"], output)
    print("write group data")
    with open("synthetic_" + file_name + "_group_data.csv", "wb") as output:
        pickle.dump(ret, output)
    return True



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass, field
from functools import singledispatch
from typing import Any, Dict, Type, Union

import numpy as np


@dataclass
class Schema:
    fields: Dict[str, Type] = field(default_factory=dict)

    def __getitem__(self, key):
        return self.fields[key]

    def get(self, key, default):
        return self.fields.get(key, default)

    def __contains__(self, key):
        return key in self.fields

    def copy(self):
        return Schema(self.fields.copy())

    def add(self, name: str, ty: Type, force=False) -> "Schema":
        if name in self and not force:
            raise KeyError(f"Can't override existing field {name!r}.")

        clone = self.copy()
        clone.fields[name] = ty
        return clone

    def union(self, name: str, ty: Type) -> "Schema":
        clone = self.copy()
        clone.fields[name] = Union[self.fields.get(name, ty), ty]
        return clone

    def remove(self, name: str, ignore_missing=False) -> "Schema":
        if name not in self and not ignore_missing:
            raise KeyError(f"Can't remove missing field {name!r}.")

        clone = self.copy()
        clone.fields.pop(name, None)
        return clone

    def pop(self, name: str) -> "Schema":
        if name not in self:
            raise KeyError(f"Can't pop missing field {name!r}.")

        clone = self.copy()
        ty = clone.fields.pop(name)
        return clone, ty

    def validate(self, entry: dict):
        return {name: ty(entry[name]) for name, ty in self.fields.items()}

    @classmethod
    def infer(cls, data):
        return cls({name: infer_type(value) for name, value in data.items()})

    @classmethod
    def dryrun(cls, pipeline):
        return pipeline.apply_schema(SchemaTracker())


@dataclass
class SchemaTracker:
    output_schema: Dict[str, Type] = field(default_factory=Schema)
    input_schema: Dict[str, Type] = field(default_factory=Schema)

    def _with_missing(self, key, Ty=Any):
        if key in self.output_schema:
            return self, self.output_schema[key]

        Ty_ = self.input_schema.get(key, Ty)
        combined = Union[Ty, Ty_]
        clone = self.copy()

        clone.input_schema.fields[key] = combined
        return clone, combined

    def copy(self):
        return SchemaTracker(
            self.output_schema.copy(), self.input_schema.copy()
        )

    def add(self, name: str, ty: Type, force=False) -> "Schema":
        if name in self.output_schema and not force:
            raise KeyError(f"Can't override existing field {name!r}.")

        elif name in self.input_schema and not force:
            raise KeyError(f"Overwrite shadowed field {name!r}.")

        clone = self.copy()
        clone.output_schema.fields[name] = ty
        return clone

    def union(self, name: str, ty: Type) -> "Schema":
        clone = self.copy()
        clone.output_schema.fields[name] = Union[
            self.output_schema.get(name, ty), ty
        ]
        return clone

    def remove(self, name: str, ignore_missing=False) -> "Schema":
        schema, _ty = self._with_missing(name)
        schema.output_schema.fields.pop(name, None)
        return schema

    def pop(self, name: str) -> "Schema":
        schema, ty = self._with_missing(name)

        schema.output_schema.fields.pop(name, None)
        return schema, ty


@singledispatch
def infer_type(val: Any):
    return type(val)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import numpy as np

import torch
import torch.nn as nn

from pts.modules import FeatureEmbedder as BaseFeatureEmbedder


class FeatureProjector(nn.Module):
    def __init__(
        self,
        feature_dims: List[int],
        embedding_dims: List[int],
    ):
        super().__init__()

        self.__num_features = len(feature_dims)
        if self.__num_features > 1:
            self.feature_slices = (
                feature_dims[0:1] + np.cumsum(feature_dims)[:-1].tolist()
            )
        else:
            self.feature_slices = feature_dims
        self.feature_dims = feature_dims

        self._projector = nn.ModuleList(
            [
                nn.Linear(in_features=in_feature, out_features=out_features)
                for in_feature, out_features in zip(
                    self.feature_dims, embedding_dims
                )
            ]
        )

    def forward(self, features: torch.Tensor) -> List[torch.Tensor]:
        if self.__num_features > 1:
            real_feature_slices = torch.tensor_split(
                features, self.feature_slices[1:], dim=-1
            )
        else:
            real_feature_slices = [features]

        return [
            proj(real_feature_slice)
            for proj, real_feature_slice in zip(
                self._projector, real_feature_slices
            )
        ]


class FeatureEmbedder(BaseFeatureEmbedder):
    def forward(self, features: torch.Tensor) -> List[torch.Tensor]:
        concat_features = super(FeatureEmbedder, self).forward(
            features=features
        )

        if self.__num_features > 1:
            features = torch.chunk(
                concat_features, self.__num_features, dim=-1
            )
        else:
            features = [concat_features]

        return features


class GatedLinearUnit(nn.Module):
    def __init__(self, dim: int = -1, nonlinear: bool = True):
        super().__init__()
        self.dim = dim
        self.nonlinear = nonlinear

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        val, gate = torch.chunk(x, 2, dim=self.dim)
        if self.nonlinear:
            val = torch.tanh(val)
        return torch.sigmoid(gate) * val


class GatedResidualNetwork(nn.Module):
    def __init__(
        self,
        d_hidden: int,
        d_input: Optional[int] = None,
        d_output: Optional[int] = None,
        d_static: Optional[int] = None,
        dropout: float = 0.0,
    ):
        super().__init__()

        d_input = d_input or d_hidden
        d_static = d_static or 0
        if d_output is None:
            d_output = d_input
            self.add_skip = False
        else:
            if d_output != d_input:
                self.add_skip = True
                self.skip_proj = nn.Linear(
                    in_features=d_input, out_features=d_output
                )
            else:
                self.add_skip = False

        self.mlp = nn.Sequential(
            nn.Linear(in_features=d_input + d_static, out_features=d_hidden),
            nn.ELU(),
            nn.Linear(in_features=d_hidden, out_features=d_hidden),
            nn.Dropout(p=dropout),
            nn.Linear(in_features=d_hidden, out_features=d_output * 2),
            GatedLinearUnit(nonlinear=False),
        )

        self.lnorm = nn.LayerNorm(d_output)

    def forward(
        self, x: torch.Tensor, c: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        if self.add_skip:
            skip = self.skip_proj(x)
        else:
            skip = x

        if c is not None:
            x = torch.cat((x, c), dim=-1)
        x = self.mlp(x)
        x = self.lnorm(x + skip)
        return x


class VariableSelectionNetwork(nn.Module):
    def __init__(
        self,
        d_hidden: int,
        n_vars: int,
        dropout: float = 0.0,
        add_static: bool = False,
    ):
        super().__init__()
        self.weight_network = GatedResidualNetwork(
            d_hidden=d_hidden,
            d_input=d_hidden * n_vars,
            d_output=n_vars,
            d_static=d_hidden if add_static else None,
            dropout=dropout,
        )

        self.variable_network = nn.ModuleList(
            [
                GatedResidualNetwork(d_hidden=d_hidden, dropout=dropout)
                for _ in range(n_vars)
            ]
        )

    def forward(
        self,
        variables: List[torch.Tensor],
        static: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        flatten = torch.cat(variables, dim=-1)
        if static is not None:
            static = static.expand_as(variables[0])
        weight = self.weight_network(flatten, static)
        weight = torch.softmax(weight.unsqueeze(-2), dim=-1)

        var_encodings = [
            net(var) for var, net in zip(variables, self.variable_network)
        ]
        var_encodings = torch.stack(var_encodings, dim=-1)

        var_encodings = torch.sum(var_encodings * weight, dim=-1)

        return var_encodings, weight


class TemporalFusionEncoder(nn.Module):
    def __init__(
        self,
        d_input: int,
        d_hidden: int,
    ):
        super().__init__()

        self.encoder_lstm = nn.LSTM(
            input_size=d_input, hidden_size=d_hidden, batch_first=True
        )
        self.decoder_lstm = nn.LSTM(
            input_size=d_input, hidden_size=d_hidden, batch_first=True
        )

        self.gate = nn.Sequential(
            nn.Linear(in_features=d_hidden, out_features=d_hidden * 2),
            GatedLinearUnit(nonlinear=False),
        )
        if d_input != d_hidden:
            self.skip_proj = nn.Linear(
                in_features=d_input, out_features=d_hidden
            )
            self.add_skip = True
        else:
            self.add_skip = False

        self.lnorm = nn.LayerNorm(d_hidden)

    def forward(
        self,
        ctx_input: torch.Tensor,
        tgt_input: torch.Tensor,
        states: List[torch.Tensor],
    ):
        ctx_encodings, states = self.encoder_lstm(ctx_input, states)

        tgt_encodings, _ = self.decoder_lstm(tgt_input, states)

        encodings = torch.cat((ctx_encodings, tgt_encodings), dim=1)
        skip = torch.cat((ctx_input, tgt_input), dim=1)
        if self.add_skip:
            skip = self.skip_proj(skip)
        encodings = self.gate(encodings)
        encodings = self.lnorm(skip + encodings)
        return encodings


class TemporalFusionDecoder(nn.Module):
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        d_hidden: int,
        d_var: int,
        n_head: int,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.context_length = context_length
        self.prediction_length = prediction_length

        self.enrich = GatedResidualNetwork(
            d_hidden=d_hidden,
            d_static=d_var,
            dropout=dropout,
        )

        self.attention = nn.MultiheadAttention(
            embed_dim=d_hidden, num_heads=n_head, dropout=dropout
        )

        self.att_net = nn.Sequential(
            nn.Linear(in_features=d_hidden, out_features=d_hidden * 2),
            GatedLinearUnit(nonlinear=False),
        )
        self.att_lnorm = nn.LayerNorm(d_hidden)

        self.ff_net = nn.Sequential(
            GatedResidualNetwork(d_hidden=d_hidden, dropout=dropout),
            nn.Linear(in_features=d_hidden, out_features=d_hidden * 2),
            GatedLinearUnit(nonlinear=False),
        )
        self.ff_lnorm = nn.LayerNorm(d_hidden)

        self.register_buffer(
            "attn_mask",
            self._generate_subsequent_mask(
                prediction_length, prediction_length + context_length
            ),
        )

    @staticmethod
    def _generate_subsequent_mask(
        target_length: int, source_length: int
    ) -> torch.Tensor:
        mask = (
            torch.triu(torch.ones(source_length, target_length)) == 1
        ).transpose(0, 1)
        mask = (
            mask.float()
            .masked_fill(mask == 0, float("-inf"))
            .masked_fill(mask == 1, float(0.0))
        )
        return mask

    def forward(
        self, x: torch.Tensor, static: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        static = static.repeat(
            (1, self.context_length + self.prediction_length, 1)
        )

        skip = x[:, self.context_length :, ...]
        x = self.enrich(x, static)

        mask_pad = torch.ones_like(mask)[:, 0:1, ...]
        mask_pad = mask_pad.repeat((1, self.prediction_length))
        key_padding_mask = torch.cat((mask, mask_pad), dim=1).bool()

        query_key_value = x.permute(1, 0, 2)

        attn_output, _ = self.attention(
            query=query_key_value[-self.prediction_length :, ...],
            key=query_key_value,
            value=query_key_value,
            # key_padding_mask=key_padding_mask, # does not work on GPU :-(
            attn_mask=self.attn_mask,
        )
        att = self.att_net(attn_output.permute(1, 0, 2))

        x = x[:, self.context_length :, ...]
        x = self.att_lnorm(x + att)
        x = self.ff_net(x)
        x = self.ff_lnorm(x + skip)

        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import List
import math

import torch as pt
from torch import Tensor, nn


class GeLU(nn.Module):
    """
    Gaussian error Linear Unit.

    .. math::
        y = 1/2 * x * (1 + tanh(\sqrt{2/pi} * (x + 0.044715*x^3)))
    """

    def forward(self, x: Tensor) -> Tensor:
        return (
            0.5
            * x
            * (
                1
                + pt.tanh(
                    math.sqrt(2 / math.pi) * (x + 0.044715 * pt.pow(x, 3))
                )
            )
        )


class Swish(nn.Sigmoid):
    """
    Swish activation by google https://arxiv.org/pdf/1710.05941v1.pdf.

    .. math::
        y = \sigma(x) * x
    """

    def forward(self, x: Tensor) -> Tensor:
        super(Swish, self).forward(x) * x


class PositiveSoftplus(nn.Softplus):
    """
    Softplus function that ensures a strictly positive activation.

    Parameters
    ----------
    margin : float
        the minimum value of activation. when =0, same as vanilla softplus
    beta: float
    threshold: float
    """

    def __init__(
        self, margin: float = 0.0, beta: float = 1.0, threshold: float = 20.0
    ):
        super(PositiveSoftplus, self).__init__(beta, threshold)
        assert margin >= 0.0
        self.margin = margin

    def forward(self, x: Tensor) -> Tensor:
        return (
            super(PositiveSoftplus, self).forward(x - self.margin)
            + self.margin
        )


class GatedLinearUnit(nn.Module):
    """
    Gated Linear Unit activation proposed by
    https://arxiv.org/pdf/1612.08083.pdf.

    Parameters
    ----------
    dim : int
        the dimension to split gate and activation
    nonlinear: bool (default True)
        if True apply a tanh to activation before applying gate
    """

    def __init__(self, dim: int = -1, nonlinear: bool = True):
        super(GatedLinearUnit, self).__init__()
        self.dim = dim
        self.nonlinear = nonlinear

    def forward(self, x: Tensor) -> Tensor:
        if x.size(self.dim) % 2 > 0:
            raise ValueError("The specified dimension must be even")
        val, gate = x.chunk(2, dim=self.dim)
        if self.nonlinear:
            val = pt.tanh(val)
        gate = pt.sigmoid(gate)
        return val * gate



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import List, Tuple

import torch as pt
from torch import Tensor
from torch import nn
from torch.distributions import Distribution

from .activations import PositiveSoftplus


def distribution_cat(distributions: List[Distribution], dim=0) -> Distribution:
    args = list(distributions[0].arg_constraints.keys())
    if "probs" in args and "logits" in args:
        try:
            _ = getattr(distributions[0], "probs")
            args.remove("logits")
        except AttributeError:
            args.remove("probs")
    concat_args = {
        arg: pt.cat([getattr(d, arg) for d in distributions], dim=dim)
        for arg in args
    }
    return type(distributions[0])(**concat_args)


class GaussianLayer(nn.Module):
    def __init__(self, d_hidden: int, d_data: int):
        super().__init__()
        self.mean = nn.Linear(d_hidden, d_data)
        self.var = nn.Sequential(
            nn.Linear(d_hidden, d_data), PositiveSoftplus(1e-3)
        )

    def forward(self, hidden: Tensor) -> Tuple[Tensor, Tensor]:
        mu = self.mean(hidden)
        sigma = self.var(hidden)
        return mu, sigma



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import pickle
import torch
from tqdm.auto import tqdm
import warnings

warnings.filterwarnings("ignore")
from datasets import DATASETS, get_dataset
import os
from utils import (
    Metrics,
    Params,
    calc_loss,
    load_pickle,
    change_device,
    PREDICTION_INPUT_NAMES,
    smoothed_inference,
)
from src.gluonts.model.predictor import Predictor
from pathlib import Path
from read_pickle import create_table


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "model_path", type=str, help="path to model checkpoint"
    )
    parser.add_argument(
        "--attack_result", type=str, help="path to the attack result"
    )
    parser.add_argument(
        "--dataset", type=str, default="electricity", choices=DATASETS
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cuda" if torch.cuda.is_available() else "cpu",
        help="cpu or cuda w/ number specified",
    )
    parser.add_argument(
        "--attack_params_path",
        type=str,
        required=True,
        help="path to json file containing attack parameters",
    )
    parser.add_argument("--tol", type=float, default=2.0)
    parser.add_argument(
        "--rs", action="store_true", help="whether to use randomized smoothing"
    )
    args = parser.parse_args()

    device = args.device
    path = Path(args.model_path)
    predictor = Predictor.deserialize(path)

    params = Params(json_path=args.attack_params_path)

    filename = args.attack_result

    attack_data = load_pickle("./attack_results_sparse/" + filename)
    attack_data = list(attack_data)

    ds = get_dataset(args.dataset, predictor.prediction_net.target_dim)

    tolerance = params.tolerance
    attack_idx = params.attack_idx
    sigmas = params.sigmas
    target_items = params.target_items
    attack_items = params.attack_items
    sparsity = params.sparsity

    net = predictor.prediction_net.to(device)
    net.eval()
    forecasts_keys = ["no attack"] + [str(k) for k in sparsity]
    forecasts = {key: [] for key in forecasts_keys}
    with torch.no_grad():
        for i in tqdm(range(len(attack_data))):
            batch = attack_data[i].batch
            batch = change_device(batch, device)
            tol = batch["past_target_cdf"].abs().max() * args.tol
            """
            Clean data, original model
            """
            inputs = dict(
                [(key, batch[key]) for key in PREDICTION_INPUT_NAMES]
            )
            if args.rs:
                outputs = smoothed_inference(
                    batch, batch["past_target_cdf"], net, params.sigma, device
                )
            else:
                outputs = net(**inputs).detach().cpu().numpy()
            forecasts["no attack"].append(outputs)

            del inputs, outputs
            torch.cuda.empty_cache()
            """
            Sparse attacked data
            """
            for s in params.sparsity:
                perturbation = attack_data[i].perturbation[s]
                perturbation = torch.clamp(perturbation, max=tol)
                batch_size = perturbation.shape[0]

                perturbation_tensor = change_device(perturbation, device)
                attacked_past_target = change_device(
                    batch["past_target_cdf"], device
                ) * (
                    1 + perturbation_tensor
                )  # perturbation: batch x context length x target dim

                attacked_inputs = dict(
                    [(key, batch[key]) for key in PREDICTION_INPUT_NAMES]
                )
                attacked_inputs["past_target_cdf"] = attacked_past_target
                if args.rs:
                    outputs = smoothed_inference(
                        batch, attacked_past_target, net, params.sigma, device
                    )
                else:
                    outputs = net(**attacked_inputs).detach().cpu().numpy()
                forecasts[str(s)].append(outputs)

                del perturbation_tensor, attacked_inputs, outputs
                torch.cuda.empty_cache()

        mse, mape, ql = calc_loss(
            attack_data,
            forecasts,
            attack_idx=attack_idx,
            target_items=target_items,
        )
        metrics = Metrics(mse=mse, mape=mape, ql=ql)

    if not args.rs:
        if not os.path.exists("./metrics_sparse"):
            os.makedirs("./metrics_sparse")
        with open("./metrics_sparse/" + filename, "wb") as outp:
            pickle.dump(metrics, outp, pickle.HIGHEST_PROTOCOL)
            pickle_path = "./metrics_sparse/" + filename
    else:
        if not os.path.exists("./metrics_sparse_rs"):
            os.makedirs("./metrics_sparse_rs")
        with open("./metrics_sparse_rs/" + filename, "wb") as outp:
            pickle.dump(metrics, outp, pickle.HIGHEST_PROTOCOL)
            pickle_path = "./metrics_sparse_rs/" + filename

    print(pickle_path)
    create_table(pickle_path)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
from sklearn.model_selection import train_test_split
from math import isclose
import numpy as np


FREQ_M1_M3 = ["monthly", "quarterly", "yearly"]
FREQ_M4 = ["daily", "weekly", "monthly", "quarterly", "yearly"]

CATEGORIES_M1 = [
    "micro1",
    "micro2",
    "micro3",
    "indust",
    "macro1",
    "macro2",
    "demogr",
]

CATEGORIES_M3_M4 = [
    "micro",
    "industry",
    "macro",
    "finance",
    "demographic",
]


M1 = [
    "m1_" + freq + "_" + cat
    for freq, cat in itertools.product(FREQ_M1_M3, CATEGORIES_M1)
]
M3 = [
    "m3_" + freq + "_" + cat
    for freq, cat in itertools.product(FREQ_M1_M3, CATEGORIES_M3_M4)
]
M4 = [
    "m4_" + freq + "_" + cat
    for freq, cat in itertools.product(FREQ_M4, CATEGORIES_M3_M4)
]

DATASETS_SINGLE = [
    # "exchange_rate",
    "exchange_rate_nips",
    # "solar-energy",
    "solar_nips",
    # "electricity",
    "electricity_nips",
    # "traffic",
    "traffic_nips",
    "wiki-rolling_nips",  # is a very challenging dataset
    "taxi_30min",
    "kaggle_web_traffic_without_missing",
    "kaggle_web_traffic_weekly",
    "nn5_daily_without_missing",  # http://www.neural-forecasting-competition.com/NN5/  (ATM cash withdrawals)
    "nn5_weekly",
    "tourism_monthly",  # no need to split, https://robjhyndman.com/papers/forecompijf.pdf
    "tourism_quarterly",
    "tourism_yearly",
    "cif_2016",  # https://irafm.osu.cz/cif/main.php?c=Static&page=download ,
    # https://zenodo.org/record/3904073#.YeE6ZMYo8UE , 72 time series, 24 real, 48 artificial, all monthly, banking domain
    "london_smart_meters_without_missing",
    "wind_farms_without_missing",
    # "car_parts_without_missing",  # intermittent time series
    # "dominick",  # This dataset contains 115704 weekly time series representing the profit of individual stock keeping units from a retailer.
    # https://www.chicagobooth.edu/research/kilts/datasets/dominicks
    # dominick has intermittency and other problems
    # TODO: Could this be clustered? (as step 2, it is already from one category only)
    "fred_md",  # https://s3.amazonaws.com/real.stlouisfed.org/wp/2015/2015-012.pdf , too mixed up?
    "pedestrian_counts",  # hourly pedestrian counts captured from 66 sensors in Melbourne city starting from May 2009
    "hospital",  # https://zenodo.org/record/4656014#.YeFBT8Yo8UE
    # This dataset contains 767 monthly time series that represent the patient counts related to medical products from January 2000 to December 2006. It was extracted from R expsmooth package.
    "covid_deaths",  # https://zenodo.org/record/4656009#.YeFBq8Yo8UE
    # This dataset contains 266 daily time series that represent the COVID-19 deaths in a set of countries and states from 22/01/2020 to 20/08/2020. It was extracted from the Johns Hopkins repository.
    "kdd_cup_2018_without_missing",  # https://zenodo.org/record/4656756#.YeFB7cYo8UE
    # This dataset was used in the KDD Cup 2018 forecasting competition.
    # It contains long hourly time series representing the air quality levels
    # in 59 stations in 2 cities: Beijing (35 stations) and London (24 stations) from 01/01/2017 to 31/03/2018.
    # TODO: could be clustered by cities
    "weather",  # https://zenodo.org/record/4654822#.YeFCXcYo8UE
    # 3010 daily time series representing the variations of four weather variables: rain, mintemp, maxtemp and solar radiation, measured at the weather stations in Australia
    # TODO: could be clustered by variable (as step 2, it is already from one category only)
    # "m5",  # intermittent time series, see https://mofc.unic.ac.cy/m5-competition/, Walmart product data, could be splitted by product or store
]

DATASETS_FULL = DATASETS_SINGLE + M1 + M3 + M4

# The excluded datasets contain only time series that are shorter than 3 predictions lengths.
# In the current setting we filter these time series.
DATASETS_FILTERED = [
    ds
    for ds in DATASETS_FULL
    if ds not in ["m1_yearly_macro2", "m3_yearly_micro", "m3_yearly_macro"]
]


def sample_datasets(
    n_folds: int,
    train_size: float = 0.7,
    val_size: float = 0.2,
    test_size: float = 0.1,
    seed=42,
):
    assert isclose(
        train_size + val_size + test_size, 1.0
    ), "sizes need to add up to 1"
    random_state = np.random.RandomState(seed)
    folds = []
    for _ in range(n_folds):
        train_split, test_split = train_test_split(
            DATASETS_FILTERED,
            train_size=train_size,
            random_state=random_state.randint(low=0, high=10000),
        )
        rest = 1 - train_size
        val_split, test_split = train_test_split(
            test_split,
            test_size=test_size / rest,
            random_state=random_state.randint(low=0, high=10000),
        )
        folds.append((train_split, val_split, test_split))
        assert not any(
            [
                set(train_split) & (set(val_split)),
                set(train_split) & set(test_split),
                set(val_split) & set(test_split),
            ]
        ), "Splits should not intersect!"
    return folds



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
import json
from dataclasses import dataclass
from typing import Tuple, Optional, List
from pathlib import Path
from src.gluonts.dataset.common import FileDataset, MetaData
from src.gluonts.dataset.field_names import FieldName
import torch
import numpy as np
import numpy.ma as ma
import shutil

from meta.data.dataset import TimeSeries, TimeSeriesDataset
from meta.common import tensor_to_np


@dataclass
class DatasetSplits:
    """
    The dataset splits provide train, validation and test data for a particular
    dataset.

    Calling any of the functions here, is a noop. Data is only loaded once a
    particular representation of the data is accessed.
    """

    _metadata: MetaData
    _directory: Path
    _dataset_name: str
    _prediction_length: int
    _standardize: bool

    def train(self, val: bool = True, name: str = "") -> DatasetSplit:
        """
        Returns the train data for the dataset.

        Args:
            val: Whether validation data is used. If not, this returns the validation data, i.e.
                the same time series that are longer by the prediction length.
            name: A specific dataset name. Should only be used for the cheat data module.
        """
        if val:
            split_name = f"train_{name}" if name else "train"
        else:
            split_name = f"val_{name}" if name else "val"

        return DatasetSplit(
            self._metadata,
            self._directory,
            split_name,
            self._dataset_name,
            _standardize=self._standardize,
        )

    def val(self, name: str = "") -> DatasetSplit:
        """
        Returns the validation data for the dataset.

        This is the same as :meth:`train(False)`.
        """
        return DatasetSplit(
            self._metadata,
            self._directory,
            f"val_{name}" if name else "val",
            self._dataset_name,
            self._standardize,
            self._prediction_length,
        )

    def test(self, name: str = "") -> DatasetSplit:
        """
        Returns the test data for the dataset.
        """
        return DatasetSplit(
            self._metadata,
            self._directory,
            f"test_{name}" if name else "test",
            self._dataset_name,
            self._standardize,
            self._prediction_length,
        )


@dataclass
class DatasetSplit:
    """
    A dataset split provides all the representations for a particular split
    (train/val/test) of a dataset.
    """

    _metadata: MetaData
    _directory: Path
    _split: str
    _dataset_name: str
    _standardize: bool
    _prediction_length: Optional[int] = None

    def data(self, evaluation: bool = False) -> TimeSeriesDataset:
        """
        Returns a time series dataset for the dataset split.

        This loads the associated JSON file and is, thus, potentially slow.
        """
        gluonts = FileDataset(
            self._directory / self._split, freq=self._metadata.freq
        )
        series = [
            TimeSeries(
                dataset_name=self._dataset_name,
                values=torch.from_numpy(item[FieldName.TARGET]).unsqueeze(-1),
                item_id=item[FieldName.ITEM_ID],
                start_date=item[FieldName.START],
            )
            for item in gluonts
        ]
        dataset = TimeSeriesDataset(
            series=series,
            prediction_length=self._metadata.prediction_length,
            freq=self._metadata.freq,
            # do not standardize evaluation dataset
            standardize=self._standardize and not evaluation,
        )
        return dataset

    def support_set(self) -> List[List[TimeSeries]]:
        support_set_path = self._directory / self._split / ".support_set.json"
        assert (
            support_set_path.exists()
        ), "No support set has been saved for this dataset."
        support_set = []
        with open(support_set_path) as json_file:
            for line in json_file:
                series = [
                    TimeSeries(
                        dataset_name=self._dataset_name,
                        values=torch.FloatTensor(s["target"]).unsqueeze(-1),
                        item_id=s["item_id"],
                        start_date=s["start"],
                    )
                    for s in json.loads(line)
                ]
                support_set.append(series)
        return support_set

    def evaluation(self) -> EvaluationDataset:
        """
        Returns the NumPy arrays that are used to perform evaluation.
        """
        if self._split == "train":
            raise ValueError(
                "training data does not provide an evaluation dataset"
            )

        base = self._directory / "numpy" / self._split
        return EvaluationDataset(
            np.load(base / "future_data.npy")[..., : self._prediction_length],
            ma.MaskedArray(
                np.load(base / "past_data.npy"),
                mask=np.load(base / "past_mask.npy"),
            ),
        )

    def prepare(self) -> None:
        """
        Prepares all required representations provided that the GluonTS dataset
        is already generated.
        """
        target = self._directory / "numpy" / self._split
        if self._split == "train":
            if target.exists():
                shutil.rmtree(target)
            return

        if target.exists():
            if (
                (target / "future_data.npy").exists()
                and (target / "past_data.npy").exists()
                and (target / "past_mask.npy").exists()
            ):
                return
            shutil.rmtree(target)

        target.mkdir(parents=True)
        # Evaluation dataset is not standardized
        future, past = _generate_evaluation_dataset(
            self.data(evaluation=True), self._metadata.prediction_length
        )

        np.save(target / "future_data.npy", future)
        np.save(target / "past_data.npy", past.data)
        np.save(target / "past_mask.npy", past.mask)


@dataclass
class EvaluationDataset:
    """
    The evaluation dataset is a simple dataset representation that contains a
    two-dimensional array of future values as well as a two-dimensional
    (masked) array of the past values that a model sees during training.

    This representation is very efficient for evaluation.
    """

    future: np.ndarray
    past: ma.MaskedArray


# -------------------------------------------------------------------------------------------------


def _generate_evaluation_dataset(
    dataset: TimeSeriesDataset, prediction_length: int
) -> Tuple[np.ndarray, ma.MaskedArray]:
    # Extract data from all the values in the dataset
    pasts = []
    past_lengths = []
    predictions = []
    for item in dataset:
        target = tensor_to_np(item.values).squeeze()
        pasts.append(target[:-prediction_length])
        past_lengths.append(target.shape[0] - prediction_length)
        predictions.append(target[-prediction_length:])

    # Compute masked past values as well as (non-masked) future values
    max_len = np.max(past_lengths)
    values = np.empty((len(dataset), max_len))
    mask = np.ones((len(dataset), max_len), dtype=bool)
    for i in range(len(dataset)):
        values[i, : past_lengths[i]] = pasts[i]
        mask[i, : past_lengths[i]] = False

    y_past = ma.masked_array(values, mask=mask)
    y_true = np.stack(predictions)
    return y_true, y_past



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Union
import pandas as pd
from tsbench.config import Config, ModelConfig
from tsbench.evaluations.metrics import Performance
from tsbench.evaluations.tracking import ModelTracker


class ConfigAnalyzer:
    """
    The config analyzer evaluates a single model configuration by obtaining its
    performance from a tracker across datasets and compiling it into a
    dataframe.
    """

    def __init__(self, tracker: ModelTracker):
        """
        Args:
            tracker: The tracker which is used to obtain performances.
        """
        self.tracker = tracker
        self.datasets = {
            c.dataset for c in self.tracker.get_evaluations().configurations
        }

    def run(
        self, model_config: Union[ModelConfig, Dict[str, ModelConfig]]
    ) -> pd.DataFrame:
        """
        Runs the evaluation, providing a configuration's performances for all
        datasets.

        Args:
            model_config: A single configuration for which to obtain performances or a mapping from
                dataset names to model configurations.

        Returns:
            The metrics on individual datasets.
        """
        results = []
        for dataset in self.datasets:
            # Construct the config
            if isinstance(model_config, dict):
                config = Config(model_config[dataset.name()], dataset)
            else:
                config = Config(model_config, dataset)

            # Get the performance and append to results
            performance = self.tracker.get_performance(config)
            df = Performance.to_dataframe([performance]).assign(
                test_dataset=dataset.name()
            )
            results.append(df)

        return pd.concat(results).set_index("test_dataset")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from functools import partial
from itertools import islice
from pathlib import Path

import pytest
import pandas as pd
import numpy as np
from lightning import seed_everything

from src.gluonts.dataset.repository import get_dataset
from src.gluonts.model.predictor import Predictor
from src.gluonts.torch.distributions import QuantileOutput, StudentTOutput
from src.gluonts.torch.model.deepar import DeepAREstimator
from src.gluonts.torch.model.deep_npts import (
    DeepNPTSEstimator,
    DeepNPTSNetworkDiscrete,
    DeepNPTSNetworkSmooth,
)
from src.gluonts.torch.model.forecast import DistributionForecast
from src.gluonts.torch.model.mqf2 import MQF2MultiHorizonEstimator
from src.gluonts.torch.model.simple_feedforward import SimpleFeedForwardEstimator
from src.gluonts.torch.model.d_linear import DLinearEstimator
from src.gluonts.torch.model.patch_tst import PatchTSTEstimator
from src.gluonts.torch.model.tide import TiDEEstimator
from src.gluonts.torch.model.lag_tst import LagTSTEstimator
from src.gluonts.torch.model.tft import TemporalFusionTransformerEstimator
from src.gluonts.torch.model.wavenet import WaveNetEstimator
from src.gluonts.torch.distributions import ImplicitQuantileNetworkOutput


@pytest.mark.parametrize(
    "estimator_constructor",
    [
        lambda dataset: DeepAREstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=StudentTOutput(beta=0.1),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
            scaling=False,
        ),
        lambda dataset: DeepAREstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            context_length=1,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
            scaling=False,
        ),
        lambda dataset: MQF2MultiHorizonEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: SimpleFeedForwardEstimator(
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: SimpleFeedForwardEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DLinearEstimator(
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DLinearEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: LagTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            freq=dataset.metadata.freq,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TemporalFusionTransformerEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=StudentTOutput(),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TemporalFusionTransformerEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DeepNPTSEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            context_length=2 * dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda dataset: PatchTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: PatchTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TiDEEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TiDEEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: WaveNetEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
    ],
)
@pytest.mark.parametrize("use_validation_data", [False, True])
def test_estimator_constant_dataset(
    estimator_constructor, use_validation_data: bool
):
    constant = get_dataset("constant")

    estimator = estimator_constructor(constant)

    if use_validation_data:
        predictor = estimator.train(
            training_data=constant.train,
            validation_data=constant.train,
        )
    else:
        predictor = estimator.train(
            training_data=constant.train,
        )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    assert predictor == predictor_copy

    forecasts = predictor_copy.predict(constant.test)

    for f in islice(forecasts, 5):
        if isinstance(f, DistributionForecast):
            f = f.to_sample_forecast()
        f.mean


@pytest.mark.parametrize(
    "estimator_constructor",
    [
        lambda freq, prediction_length: DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            distr_output=StudentTOutput(beta=0.1),
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
            distr_output=ImplicitQuantileNetworkOutput(),
        ),
        lambda freq, prediction_length: TiDEEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: MQF2MultiHorizonEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: TemporalFusionTransformerEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            dynamic_dims=[3],
            static_dims=[1],
            static_cardinalities=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=True,
            network_type=partial(DeepNPTSNetworkDiscrete, use_softmax=True),
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling=None,
            dropout_rate=0.0,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=False,
            network_type=partial(DeepNPTSNetworkDiscrete, use_softmax=False),
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling="min_max_scaling",
            dropout_rate=0.0,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=True,
            network_type=DeepNPTSNetworkSmooth,
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling="standard_normal_scaling",
            dropout_rate=0.1,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: PatchTSTEstimator(
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            num_feat_dynamic_real=3,
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: PatchTSTEstimator(
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            num_feat_dynamic_real=3,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: WaveNetEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
    ],
)
def test_estimator_with_features(estimator_constructor):
    seed_everything(42)
    freq = "1h"
    prediction_length = 12

    training_dataset = [
        {
            "start": pd.Period("2021-01-01 00:00:00", freq=freq),
            "target": np.ones(200, dtype=np.float32),
            "feat_static_cat": np.array([0, 1], dtype=np.float32),
            "feat_static_real": np.array([42.0], dtype=np.float32),
            "feat_dynamic_real": np.ones((3, 200), dtype=np.float32),
            "__unused__": np.ones(3, dtype=np.float32),
        },
        {
            "start": pd.Period("2021-02-01 00:00:00", freq=freq),
            "target": np.ones(100, dtype=np.float32),
            "feat_static_cat": np.array([1, 0], dtype=np.float32),
            "feat_static_real": np.array([1.0], dtype=np.float32),
            "feat_dynamic_real": np.ones((3, 100), dtype=np.float32),
            "__unused__": np.ones(5, dtype=np.float32),
        },
    ]

    prediction_dataset = [
        {
            "start": pd.Period("2021-01-01 00:00:00", freq=freq),
            "target": np.ones(200, dtype=np.float32),
            "feat_static_cat": np.array([0, 1], dtype=np.float32),
            "feat_static_real": np.array([42.0], dtype=np.float32),
            "feat_dynamic_real": np.ones(
                (3, 200 + prediction_length), dtype=np.float32
            ),
            "__unused__": np.ones(3, dtype=np.float32),
        },
        {
            "start": pd.Period("2021-02-01 00:00:00", freq=freq),
            "target": np.ones(100, dtype=np.float32),
            "feat_static_cat": np.array([1, 0], dtype=np.float32),
            "feat_static_real": np.array([1.0], dtype=np.float32),
            "feat_dynamic_real": np.ones(
                (3, 100 + prediction_length), dtype=np.float32
            ),
            "__unused__": np.ones(5, dtype=np.float32),
        },
    ]

    estimator = estimator_constructor(freq, prediction_length)

    predictor = estimator.train(
        training_data=training_dataset,
        validation_data=training_dataset,
    )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    assert predictor == predictor_copy

    forecasts = predictor_copy.predict(prediction_dataset)

    for f in islice(forecasts, 5):
        f.mean



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from math import pi

import torch
from torch import nn as nn


class ImplicitQuantileModule(nn.Module):
    """
    See arXiv: 1806.06923 This module, in combination with quantile loss,
    learns how to generate the quantile of the distribution of the target.

    A quantile value, tau, is randomly generated with a Uniform([0, 1])). This
    quantile value is embedded in this module and also passed to the quantile
    loss: this should force the model to learn the appropriate quantile.
    """

    def __init__(self, in_features, output_domain_cls):
        super(ImplicitQuantileModule, self).__init__()
        self.in_features = in_features
        self.quantile_layer = QuantileLayer(in_features)
        self.output_layer = nn.Sequential(
            nn.Linear(in_features, in_features),
            nn.Softplus(),
            nn.Linear(in_features, 1),
            output_domain_cls(),
        )

    def forward(self, input_data, tau):
        embedded_tau = self.quantile_layer(tau)
        new_input_data = input_data * (
            torch.ones_like(embedded_tau) + embedded_tau
        )
        return self.output_layer(new_input_data).squeeze(-1)


class QuantileLayer(nn.Module):
    """
    Define quantile embedding layer.

    i.e. phi in the IQN paper (arXiv: 1806.06923)
    """

    def __init__(self, num_output):
        super(QuantileLayer, self).__init__()
        self.n_cos_embedding = 64
        self.num_output = num_output
        self.output_layer = nn.Sequential(
            nn.Linear(self.n_cos_embedding, self.n_cos_embedding),
            nn.PReLU(),
            nn.Linear(self.n_cos_embedding, num_output),
        )

    def forward(self, tau):
        cos_embedded_tau = self.cos_embed(tau)
        final_output = self.output_layer(cos_embedded_tau)
        return final_output

    def cos_embed(self, tau):
        integers = torch.repeat_interleave(
            torch.arange(0, self.n_cos_embedding).unsqueeze(dim=0),
            repeats=tau.shape[-1],
            dim=0,
        ).to(tau.device)
        return torch.cos(pi * tau.unsqueeze(dim=-1) * integers)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Type

from src.gluonts import itertools as it


class Action:
    def apply(self, stream):
        raise NotImplementedError

    def apply_one(self, data):
        return list(self.apply([data]))

    def apply_schema(self, schema):
        raise NotImplementedError

    def bind(self, schema):
        return Bind(self, schema, self.apply_schema(schema))

    def requires(self):
        raise NotImplementedError

    def __add__(self, other):
        return other.__radd__(self)

    def __radd__(self, other):
        if isinstance(other, Pipeline):
            return Pipeline(other.actions + [self])

        return Pipeline([other, self])


@dataclass
class Bind(Action):
    action: Action
    input_schema: "Schema"
    output_schema: "Schema"

    def apply(self, stream):
        return self.action.apply(stream)


@dataclass
class Pipeline(Action):
    actions: List[Action]

    def apply(self, stream):
        for action in self.actions:
            stream = action.apply(stream)

        return stream

    def apply_schema(self, schema):
        for action in self.actions:
            schema = action.apply_schema(schema)

        return schema

    def __radd__(self, other):
        if isinstance(other, Pipeline):
            return Pipeline(other.actions + self.actions)

        return Pipeline([other] + self.actions)


class Filter(Action):
    def filter(self, data):
        raise NotImplementedError

    def apply_schema(self, schema):
        return schema

    def apply(self, data):
        return it.Filter(self.filter, data)


class Map(Action):
    def __call__(self, data: dict) -> dict:
        raise NotImplementedError

    def apply(self, stream):
        return it.Map(self, stream)

    def __radd__(self, other):
        if isinstance(other, Map):
            if isinstance(other, MapPipeline):
                return MapPipeline(other.actions + [self])
            return MapPipeline([other, self])

        return Action.__radd__(self, other)


@dataclass
class MapPipeline(Pipeline, Map):
    def __call__(self, data):
        for result in self.apply_one(data):
            return result

    def __radd__(self, other):
        if isinstance(other, MapPipeline):
            return MapPipeline(other.actions + self.actions)
        elif isinstance(other, Map):
            return MapPipeline([other] + self.actions)
        else:
            return Pipeline.__radd__(self, other)


class Identity(Map):
    def __add__(self, other):
        return other

    def __radd__(self, other):
        return other

    def __call__(self, data):
        return data

    def apply_schema(self, schema):
        return schema


class Copy(Map):
    def __call__(self, data):
        return dict(data)

    def apply_schema(self, schema):
        return schema


@dataclass
class Set(Map):
    name: str
    value: Any

    def __call__(self, data):
        data[self.name] = self.value
        return data

    def apply_schema(self, schema):
        return schema.add(self.name, type(self.value))


@dataclass
class SetDefault(Map):
    name: str
    value: Any

    def __call__(self, data):
        data.setdefault(self.name, self.value)

        return data

    def apply_schema(self, schema):
        return schema.union(self.name, type(self.value))


@dataclass
class Update(Map):
    fields: Dict[str, Type]

    def __call__(self, data):
        data.update(self.fields)
        return data

    def apply_schema(self, schema):
        for name, value in self.fields.items():
            schema = schema.add(name, type(value))

        return schema


@dataclass
class UpdateDefault(Map):
    fields: Dict[str, Type]

    def __call__(self, data):
        result = dict(self.fields)
        result.update(data)
        return result

    def apply_schema(self, schema):
        for name, value in self.fields.items():
            schema = schema.union(name, type(value))

        return schema


@dataclass
class Remove(Map):
    name: str

    def __call__(self, data):
        del data[self.name]
        return data

    def apply_schema(self, schema):
        return schema.remove(self.name)


@dataclass
class Move(Map):
    source: str
    target: str

    def __call__(self, data):
        data[self.target] = data.pop(self.source)
        return data

    def apply_schema(self, schema):
        schema, ty = schema.pop(self.source)
        return schema.set(self.target, ty)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
import yaml
from black import click
from cli.utils import (
    explode_key_values,
    iterate_configurations,
    run_sacred_script,
)
from tsbench.constants import DEFAULT_DATA_PATH, DEFAULT_EVALUATIONS_PATH
from ._main import analysis


@analysis.command(short_help="Evaluate the performance of ensembles.")
@click.option(
    "--experiment",
    required=True,
    help=(
        "The name of the experiment under which the individual training runs"
        " are grouped."
    ),
)
@click.option(
    "--config_path",
    required=True,
    help="The local path to the configuration file defining script options.",
)
@click.option(
    "--data_path",
    type=click.Path(exists=True),
    default=DEFAULT_DATA_PATH,
    show_default=True,
    help="The path where datasets are stored.",
)
@click.option(
    "--evaluations_path",
    type=click.Path(exists=True),
    default=DEFAULT_EVALUATIONS_PATH,
    show_default=True,
    help="The path where offline evaluations are stored.",
)
@click.option(
    "--nskip",
    default=0,
    show_default=True,
    help=(
        "The number of configurations to skip. Useful if some set of"
        " experiments failed."
    ),
)
def ensemble(
    experiment: str,
    config_path: str,
    data_path: str,
    evaluations_path: str,
    nskip: int,
):
    """
    Evaluates the performance of an ensemble of best-performing models. This
    allows also to only consider models of a particular type (thus building
    hyper-ensembles).

    This call runs the Sacred script for each provided configuration
    sequentially and returns only once all runs have completed.
    """
    with Path(config_path).open("r", encoding="utf-8") as f:
        content = yaml.safe_load(f)
        configs = explode_key_values("__", content)

    for configuration in iterate_configurations(configs, nskip):
        run_sacred_script(
            "ensemble.py",
            experiment=experiment,
            data_path=data_path,
            evaluations_path=evaluations_path,
            **{k: v for k, v in configuration.items() if k != "__"},
        )
